import os
import logging

import h5py
from h5py._hl.base import with_phil

logger = logging.getLogger(__name__)


def is_bag(file_name):
    """ Determine if a file is valid BAG (False if it doesn't exist). """

    # we first check if the file is a valid hdf5 (it also checks if the file exists)
    if not h5py.is_hdf5(file_name):
        return False

    fid = h5py.File(file_name, 'r')

    try:
        fid["BAG_root"]

    except KeyError:
        return False

    return True


class File(h5py.File):
    """ Represents a BAG file (at low-level, thin wrapper around h5py). """

    def __init__(self, name, mode: str = "r", driver=None,
                 libver=None, userblock_size=None, swmr=False, **kwds):
        """
        Create a new file object.

        See the h5py user guide for a detailed explanation of the options.

        name
            Name of the file on disk.  Note: for files created with the 'core'
            driver, HDF5 still requires this be non-empty.
        driver
            Name of the driver to use.  Legal values are None (default,
            recommended), 'core', 'sec2', 'stdio', 'mpio'.
        libver
            Library version bounds.  Currently only the strings 'earliest'
            and 'latest' are defined.
        userblock
            Desired size of user block.  Only allowed when creating a new
            file (mode w, w- or x).
        swmr
            Open the file in SWMR read mode. Only used when mode = 'r'.
        Additional keywords
            Passed on to the selected file driver.
        """
        super(File, self).__init__(name=name, mode=mode, driver=driver,
                                   libver=libver, userblock_size=userblock_size, swmr=swmr, **kwds)

    def close(self):
        """ Close the file.  All open objects become invalid """
        logger.debug("closing")
        super(File, self).close()

    def flush(self):
        """ Tell the BAG library to flush its buffers. """
        logger.debug("flushing")
        super(File, self).flush()

    @with_phil
    def __repr__(self):
        if not self.id:
            logger.info("closed file")
            r = '<BAG file>\n'
            r += "  <status: closed>"
        else:
            # Filename has to be forced to Unicode if it comes back bytes
            # Mode is always a "native" string
            filename = self.filename
            if isinstance(filename, bytes):  # Can't decode fname
                filename = filename.decode('utf8', 'replace')
            r = '<BAG file "%s" (mode %s)>' % (os.path.basename(filename), self.mode)
            r += "  <status: open>\n"
            r += "  <id: %s>\n" % self.id
            r += "  <name: %s>\n" % self.name
            r += "  <driver: %s>\n" % self.driver
            r += "  <user block size: %s>\n" % self.userblock_size

        return r



import os
import logging

import numpy as np
from osgeo import gdal, osr

from hyo2.bag.meta import Meta
from hyo2.bag.helper import BAGError
from hyo2.bag.bag import BAGFile


logger = logging.getLogger(__name__)
gdal.UseExceptions()


class Uncertainty2Gdal:

    formats = {
        'ascii': ["AAIGrid", "bag.uncertainty.asc"],
        'geotiff': ["GTiff", "bag.uncertainty.tif"],
        'xyz': ["XYZ", "bag.uncertainty.xyz"],
    }

    def __init__(self, bag_uncertainty: np.ndarray, bag_meta: Meta, fmt="geotiff", out_file=None, epsg=None):
        """Export the elevation layer in one of the listed formats"""
        self.bag_unc = bag_uncertainty
        self.bag_meta = bag_meta

        # get the IN-MEMORY ogr driver
        self.mem = gdal.GetDriverByName("MEM")
        if self.mem is None:
            raise BAGError("%s driver not available.\n" % self.formats[fmt][0])
        logger.debug("format: %s" % fmt)

        # set the output file
        self.out_file = out_file
        if self.out_file is None:
            self.out_file = os.path.abspath(self.formats[fmt][1])
            logger.debug("output: %s" % self.out_file)

        if os.path.exists(self.out_file):
            os.remove(self.out_file)

        logger.debug("dtype: %s" % self.bag_unc.dtype)
        self.rst = self.mem.Create(utf8_path=self.out_file, xsize=self.bag_meta.cols, ysize=self.bag_meta.rows,
                                   bands=1, eType=gdal.GDT_Float32)
        # GDAL geo-transform refers to the top left corner of the top left pixel of the raster.
        self.rst.SetGeoTransform((self.bag_meta.sw[0] - self.bag_meta.res_x / 2.0, self.bag_meta.res_x, 0,
                                  self.bag_meta.ne[1] + self.bag_meta.res_y / 2.0, 0, -self.bag_meta.res_y))

        self.bnd = self.rst.GetRasterBand(1)
        self.bnd.WriteArray(self.bag_unc[::-1])
        self.bnd.SetNoDataValue(BAGFile.BAG_NAN)
        self.srs = osr.SpatialReference()
        if self.bag_meta.wkt_srs is not None:
            self.srs.ImportFromWkt(self.bag_meta.wkt_srs)
        else:
            logger.warning("unable to recover valid spatial reference info")
        if self.srs.IsCompound():
            self.srs.StripVertical()
        self.rst.SetProjection(self.srs.ExportToWkt())
        self.bnd.FlushCache()

        # get the required ogr driver
        self.drv = gdal.GetDriverByName(self.formats[fmt][0])

        # check if re-projection is required
        if not epsg:
            # if not, we just create a copy in the selected format
            _ = self.drv.CreateCopy(self.out_file, self.rst)
            _ = None
            self.rst = None
            return

        # we need to change projection:
        # - we create the output srs
        dst_srs = osr.SpatialReference()
        dst_srs.ImportFromEPSG(epsg)
        dst_wkt = dst_srs.ExportToWkt()

        # Call AutoCreateWarpedVRT() to fetch default values for target raster dimensions and geotransform
        tmp_ds = gdal.AutoCreateWarpedVRT(self.rst,
                                          None,  # src_wkt : left to default value --> will use the one from source
                                          dst_wkt,
                                          gdal.GRA_NearestNeighbour,
                                          0.125  # error threshold --> use same value as in gdalwarp
                                          )
        # Create the final warped raster
        _ = self.drv.CreateCopy(self.out_file, tmp_ds)
        _ = None
        self.rst = None



import os
import logging

from hyo2.abc2.lib.logging import set_logging
from hyo2.bag.bag import BAGFile
from hyo2.bag.helper import Helper
from hyo2.bag.meta import Meta

set_logging(ns_list=['hyo2.bag'])
logger = logging.getLogger(__name__)

file_bag_0 = os.path.join(Helper.samples_folder(), "bdb_01.bag")
# file_bag_0 = r"D:\google_drive\_ccom\QC Tools\data\survey\BAG Checks\ResolutionCheck\F00799_MB_2m_MLLW_1of2.bag"
# file_bag_0 = r"C:\code\cpp\bag\examples\sample-data\NAVO_data\JD211_public_Release_1-4_UTM.bag"
if os.path.exists(file_bag_0):
    logger.debug("- file_bag_0: %s" % file_bag_0)

bag_0 = BAGFile(file_bag_0, mode="r")
xml_string_no_pretty = bag_0.metadata(as_string=True, as_pretty_xml=False)
logger.debug("xml_string_no_pretty: %s\n%s" % (type(xml_string_no_pretty), xml_string_no_pretty))
xml_string_pretty = bag_0.metadata(as_string=True, as_pretty_xml=True)
logger.debug("xml_string_pretty: %s\n%s" % (type(xml_string_pretty), xml_string_pretty))
xml_bytes_no_pretty = bag_0.metadata(as_string=False, as_pretty_xml=False)
logger.debug("xml_bytes_no_pretty: %s\n%s" % (type(xml_bytes_no_pretty), xml_bytes_no_pretty))
xml_bytes_pretty = bag_0.metadata(as_string=False, as_pretty_xml=True)
logger.debug("xml_bytes_pretty: %s\n%s" % (type(xml_bytes_pretty), xml_bytes_pretty))
bag_0.populate_metadata()
logger.debug(bag_0)




# -*- coding: utf-8 -*-

import sys
import os

this = os.path.dirname(os.path.abspath(__file__))

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

sys.path.insert(0, os.path.join(this, os.pardir))

# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.

#needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.coverage',
    'sphinx.ext.imgmath',
    'sphinx.ext.viewcode',
    'sphinx.ext.coverage',
    'sphinx.ext.intersphinx',
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix of source filenames.

source_suffix = '.rst'

# The encoding of source files.

#source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = u'HydrOffice BAG'
copyright = u'2024, University of New Hampshire, Center for Coastal and Ocean Mapping'


# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = '1.3'
# The full version, including alpha/beta/rc tags.
release = '1.3.1'

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#language = None


# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:

#today = ''

# Else, today_fmt is used as the format for a strftime call.

#today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.

exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# The reST default role (used for this markup: `text`) to use for all
# documents.

#default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.

#add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).

#add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.

#show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'colorful'

# A list of ignored prefixes for module index sorting.
#modindex_common_prefix = []

# If true, keep warnings as "system message" paragraphs in the built documents.
#keep_warnings = False


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.

html_theme = 'qc4'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.

#html_theme_options = {}

# Add any paths that contain custom themes here, relative to this directory.
html_theme_path = ['_theme']

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".

#html_title = None

# A shorter title for the navigation bar.  Default is the same as html_title.

#html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.

#html_logo = None

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.

#html_favicon = None

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# Add any extra paths that contain custom files (such as robots.txt or
# .htaccess) here, relative to this directory. These files are copied
# directly to the root of the documentation.

#html_extra_path = []

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.


#html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.

html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.

html_sidebars = {
    'index': ['sidebarintro.html', 'sourcelink.html', 'searchbox.html'],
    '**': ['sidebarlogo.html', 'relations.html',
           'sourcelink.html', 'searchbox.html'],
}

# Additional templates that should be rendered to pages, maps page names to
# template names.

#html_additional_pages = {}

# If false, no module index is generated.

#html_domain_indices = True

# If false, no index is generated.

html_use_index = True

# If true, the index is split into individual pages for each letter.

#html_split_index = False

# If true, links to the reST sources are added to the pages.

#html_show_sourcelink = True

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.

html_show_sphinx = False


# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.

#html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
#html_file_suffix = None

# Output file base name for HTML help builder.
htmlhelp_basename = 'HydrOfficeBAGdoc'


# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
# The paper size ('letterpaper' or 'a4paper').

#'papersize': 'letterpaper',

# The font size ('10pt', '11pt' or '12pt').

#'pointsize': '10pt',

# Additional stuff for the LaTeX preamble.

#'preamble': '',




}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
  ('index', 'HydrOfficeBAG.tex', u'HydrOffice BAG Documentation',
   u'University of New Hampshire, Center for Coastal and Ocean Mapping', 'manual'),

]

# The name of an image file (relative to this directory) to place at the top of
# the title page.

#latex_logo = None

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.

#latex_use_parts = False

# If true, show page references after internal links.

#latex_show_pagerefs = False

# If true, show URL addresses after external links.

#latex_show_urls = False

# Documents to append as an appendix to all manuals.

#latex_appendices = []


# If false, no module index is generated.

#latex_domain_indices = True


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    ('index', 'hydrofficebag', u'HydrOffice BAG Documentation',
     [u'University of New Hampshire, Center for Coastal and Ocean Mapping'], 1)


]

# If true, show URL addresses after external links.

#man_show_urls = False


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
  ('index', 'HydrOfficeBAG', u'HydrOffice BAG Documentation',
   u'University of New Hampshire, Center for Coastal and Ocean Mapping', 'HydrOfficeBAG',
   'Library and tools for working with BAG files.',


   'Miscellaneous'),
]

# Documents to append as an appendix to all manuals.

#texinfo_appendices = []

# If false, no module index is generated.

#texinfo_domain_indices = True

# How to display URL addresses: 'footnote', 'no', or 'inline'.

#texinfo_show_urls = 'footnote'

# If true, do not generate a @detailmenu in the "Top" node's menu.

#texinfo_no_detailmenu = False




import os
import logging
from PySide2 import QtWidgets  # required by matplotlib
from matplotlib import pyplot as plt

from hyo2.abc2.lib.logging import set_logging
from hyo2.bag.bag import BAGFile
from hyo2.bag.bag import BAGError
from hyo2.bag.helper import Helper
from hyo2.bag.bbox import Bbox2Gdal

logger = logging.getLogger(__name__)
set_logging(ns_list=['hyo2.bag'])

file_bag_0 = os.path.join(Helper.samples_folder(), "bdb_01.bag")
if os.path.exists(file_bag_0):
    logger.debug("- file_bag_0: %s" % file_bag_0)

bag_0 = BAGFile(file_bag_0, mode='r')
logger.debug(bag_0)

logger.debug(type(bag_0.elevation(mask_nan=True)))
logger.debug(bag_0.elevation(mask_nan=True).shape)
logger.debug(bag_0.elevation(mask_nan=True).dtype)
# ax =plt.contourf(bag_0.elevation(mask_nan=True))
# plt.colorbar(ax)
# plt.show()

bag_meta = bag_0.populate_metadata()
logger.debug(bag_meta)

Bbox2Gdal(bag_meta, fmt="gjs", title=os.path.basename(file_bag_0))
Bbox2Gdal(bag_meta, fmt="gml", title=os.path.basename(file_bag_0))
Bbox2Gdal(bag_meta, fmt="kml", title=os.path.basename(file_bag_0))
Bbox2Gdal(bag_meta, fmt="shp", title=os.path.basename(file_bag_0))



import os
import logging

from hyo2.abc2.lib.logging import set_logging
from hyo2.bag.base import is_bag, File
from hyo2.bag.helper import Helper

logger = logging.getLogger(__name__)
set_logging(ns_list=['hyo2.bag'])

# Example that use bag.base to:
# - test for being a BAG file
# - open and read some metadata info from a BAG file

file_bag_0 = os.path.join(Helper.samples_folder(), "bdb_01.bag")
if os.path.exists(file_bag_0):
    logger.info("- file_bag_0: %s is BAG? %r" % (file_bag_0, is_bag(file_bag_0)))

file_bag_1 = os.path.join(Helper.samples_folder(), "bdb_02.bag")
if os.path.exists(file_bag_1):
    logger.info("- file_bag_1: %s is BAG? %r" % (file_bag_1, is_bag(file_bag_1)))

file_fake_0 = os.path.join(Helper.samples_folder(), "fake_00.bag")
if os.path.exists(file_fake_0):
    logger.info("- file_fake_0: %s is BAG? %r" % (file_fake_0, is_bag(file_fake_0)))

bag_0 = File(file_bag_0)
logger.info("\n%s\n" % bag_0)
logger.info("filename: %s" % bag_0.filename)
logger.info("attributes for %s: %d" % (bag_0.attrs, len(bag_0.attrs)))
logger.info("driver: %s" % bag_0.driver)
bag_0.flush()
bag_0.close()
logger.info("\n%s\n" % bag_0)

bag_1 = File(file_bag_1)
bag_1.flush()
bag_1.close()

try:
    fake_0 = File(file_fake_0)
except IOError:
    logger.info("Expected exception")






import os
import logging

logger = logging.getLogger(__name__)


class BAGError(Exception):
    """ BAG class for exceptions"""

    def __init__(self, message, *args):
        if isinstance(message, Exception):
            msg = message.args[0] if len(message.args) > 0 else ''
        else:
            msg = message

        self.message = msg
        # allow users initialize misc. arguments as any other builtin Error
        Exception.__init__(self, message, *args)


class Helper(object):
    @classmethod
    def samples_folder(cls):
        samples_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), "samples"))
        if not os.path.exists(samples_dir):
            raise BAGError("unable to find the samples folder: %s" % samples_dir)
        return samples_dir

    @classmethod
    def iso19139_folder(cls):
        iso_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), "iso19139"))
        if not os.path.exists(iso_dir):
            raise BAGError("unable to find the iso19139 folder: %s" % iso_dir)
        return iso_dir

    @classmethod
    def iso19757_3_folder(cls):
        iso_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), "iso19757-3"))
        if not os.path.exists(iso_dir):
            raise BAGError("unable to find the iso19757-3 folder: %s" % iso_dir)
        return iso_dir

    @staticmethod
    def elide(input_str, max_len=255):
        """ only in case the passed string is longer than 'max_len', it applies elision """
        if len(input_str) > max_len:
            return input_str[:max_len] + "[..]"
        return input_str



import argparse
import os
import logging

from hyo2.abc2.lib.logging import set_logging
from hyo2.bag import __version__
from hyo2.bag.bag import BAGFile, is_bag

logger = logging.getLogger(__name__)


def main():
    app_name = "bag_elevation"
    app_info = "Extraction of elevation layer from an OpenNS BAG file, using hyo2.bag r%s" % __version__

    formats = ['ascii', 'geotiff', 'xyz']

    parser = argparse.ArgumentParser(prog=app_name, description=app_info)
    parser.add_argument("bag_file", type=str, help="a valid BAG file from which to extract metadata")
    parser.add_argument("-f", "--format", help="one of the available file format: " + ", ".join(formats),
                        choices=formats, default="geotiff", metavar='')
    parser.add_argument("-o", "--output", help="the output file", type=str)
    parser.add_argument("-v", "--verbose", help="increase output verbosity", action="store_true")
    args = parser.parse_args()

    if args.verbose:
        set_logging(ns_list=['hyo2.bag'])
        logger.debug("> verbosity: ON")

        logger.debug("> input: %s" % args.bag_file)

        if args.output:
            args.output = os.path.abspath(args.output)
            logger.debug("> output: %s" % args.output)
        else:
            args.output = None
            logger.debug("> output: [default]")

        logger.debug("> format: %s" % args.format)

    if not os.path.exists(args.bag_file):
        parser.exit(1, "ERROR: the input valid does not exist: %s" % args.bag_file)

    if not is_bag(args.bag_file):
        parser.exit(1, "ERROR: the input valid does not seem a BAG file: %s" % args.bag_file)

    bf = BAGFile(args.bag_file, mode='r')
    bag_meta = None
    try:
        bag_meta = bf.populate_metadata()
    except Exception as e:
        parser.exit(1, "ERROR: issue in metadata population: %s" % e)
    bag_elevation = None
    try:
        bag_elevation = bf.elevation(mask_nan=False)
    except Exception as e:
        parser.exit(1, "ERROR: issue in elevation population: %s" % e)

    try:
        from hyo2.bag.elevation import Elevation2Gdal
        Elevation2Gdal(bag_elevation=bag_elevation, bag_meta=bag_meta, fmt=args.format, out_file=args.output)
    except Exception as e:
        parser.exit(1, "ERROR: issue in output creation: %s" % e)

    if args.verbose:
        logger.debug("> DONE")


if __name__ == "__main__":
    main()



import os
import logging
from typing import Tuple

from osgeo import osr
import numpy as np
from lxml import etree, isoschematron

from hyo2.bag.base import is_bag, File
from hyo2.bag.helper import BAGError, Helper
from hyo2.bag.meta import Meta

logger = logging.getLogger(__name__)


class BAGFile(File):
    """ Represents a BAG file. """

    _bag_root = "BAG_root"
    _bag_version_tag = "Bag Version"
    _bag_version_number = b'1.6.3'
    _bag_elevation = "BAG_root/elevation"
    _bag_elevation_min_ev = "Minimum Elevation Value"
    _bag_elevation_max_ev = "Maximum Elevation Value"
    _bag_metadata = "BAG_root/metadata"
    _bag_tracking_list = "BAG_root/tracking_list"
    _bag_tracking_list_len = "Tracking List Length"
    _bag_tracking_list_type = np.dtype([('row', np.uint32), ('col', np.uint32),
                                        ('depth', np.float32), ('uncertainty', np.float32),
                                        ('track_code', np.byte), ('list_series', np.uint16)])
    _bag_uncertainty = "BAG_root/uncertainty"
    _bag_uncertainty_min_uv = "Minimum Uncertainty Value"
    _bag_uncertainty_max_uv = "Maximum Uncertainty Value"
    _bag_elevation_solution = "BAG_root/elevation_solution"

    _bag_varres_metadata = "BAG_root/varres_metadata"
    _bag_varres_refinements = "BAG_root/varres_refinements"
    _bag_varres_tracking_list = "BAG_root/varres_tracking_list"

    BAG_NAN = 1000000

    default_metadata_file = "BAG_metadata.xml"

    def __init__(self, name, mode='r', driver=None,
                 libver=None, userblock_size=None, swmr=False, **kwds):
        """
        Create a new file object.

        See the low level bag.File for a detailed explanation of the options.
        """
        if mode is not None:
            if 'w' not in mode:
                if not is_bag(name):
                    raise BAGError("The passed file %s is not a BAG file")

        super(BAGFile, self).__init__(name=name, mode=mode, driver=driver,
                                      libver=libver, userblock_size=userblock_size, swmr=swmr, **kwds)

        self.meta = None
        self.meta_errors = list()
        self._str = None

    @classmethod
    def is_bag(cls, bag_path: str, advanced: bool = False) -> bool:
        if not advanced:
            return os.path.splitext(bag_path)[-1].lower() == ".bag"

        raise RuntimeError("Not implemented")

    @classmethod
    def is_vr(cls, bag_path: str, advanced: bool = False) -> bool:
        if not advanced:
            return BAGFile(bag_path).has_varres_refinements()

        raise RuntimeError("Not implemented")

    @classmethod
    def create_template(cls, name):
        """ create a BAG file with empty template structure """
        logger.debug("create new BAG file: %s" % name)
        try:
            new_bag = File(name, 'w')
            new_bag.create_group(cls._bag_root)
            new_bag.attrs.create(cls._bag_version_tag, cls._bag_version_number, shape=(), dtype="S5")

            elevation = new_bag.create_dataset(cls._bag_elevation, shape=(), dtype=np.float32)
            elevation.attrs.create(cls._bag_elevation_min_ev, 0.0, shape=(), dtype=np.float32)
            elevation.attrs.create(cls._bag_elevation_max_ev, 0.0, shape=(), dtype=np.float32)

            new_bag.create_dataset(cls._bag_metadata, shape=(1,), dtype="S1")

            tracking_list = new_bag.create_dataset(cls._bag_tracking_list, shape=(), dtype=cls._bag_tracking_list_type)
            tracking_list.attrs.create(cls._bag_tracking_list_len, 0, shape=(), dtype=np.uint32)

            uncertainty = new_bag.create_dataset(cls._bag_uncertainty, shape=(), dtype=np.float32)
            uncertainty.attrs.create(cls._bag_uncertainty_min_uv, 0.0, shape=(), dtype=np.float32)
            uncertainty.attrs.create(cls._bag_uncertainty_max_uv, 0.0, shape=(), dtype=np.float32)

        except (BAGError, OSError) as e:
            raise BAGError("Unable to create the BAG file %s: %s" % (name, e))

        return new_bag

    def has_bag_root(self):
        return self._bag_root in self

    def has_bag_version(self):
        return self._bag_version_tag in self[self._bag_root].attrs

    def bag_version(self) -> str:
        return self[self._bag_root].attrs[self._bag_version_tag]

    def has_metadata(self):
        return BAGFile._bag_metadata in self

    def has_elevation(self):
        return BAGFile._bag_elevation in self

    def elevation(self, mask_nan=True, row_range=None):
        """
        Return the elevation as numpy array

        mask_nan
            If True, apply a mask using the BAG nan value
        row_range
            If present, a slice of rows to read from
        """
        if row_range:
            if not isinstance(row_range, slice):
                raise BAGError("Invalid type of slice selector: %s" % type(row_range))
            if (row_range.start < 0) or (row_range.start >= self.elevation_shape()[0]) \
                    or (row_range.stop < 0) or (row_range.stop > self.elevation_shape()[0]) \
                    or (row_range.start > row_range.stop):
                raise BAGError("Invalid values for slice selector: %s" % row_range)

        if mask_nan:
            if row_range:
                el = self[BAGFile._bag_elevation][row_range]
            else:
                el = self[BAGFile._bag_elevation][:]
            mask = el == BAGFile.BAG_NAN
            el[mask] = np.nan
            return el

        if row_range:
            return self[BAGFile._bag_elevation][row_range]
        else:
            return self[BAGFile._bag_elevation][:]

    def elevation_shape(self):
        return self[BAGFile._bag_elevation].shape

    def elevation_min_max(self) -> Tuple[float, float]:
        rows, cols = self.elevation_shape()
        # logger.debug('shape: %s, %s' % (rows, cols))

        mem_row = cols * 32 / 1024 / 1024
        # mem = mem_row * rows
        # logger.debug('estimated memory: %.1f MB' % mem)
        chunk_size = 8096
        chunk_rows = int(chunk_size / mem_row) + 1
        # logger.debug('nr of rows per chunk: %s' % chunk_rows)

        elv_min = np.nan
        elv_max = np.nan
        for start in range(0, rows, chunk_rows):
            stop = start + chunk_rows
            if stop > rows:
                stop = rows
            # logger.debug('slice: %s-%s' % (start, stop))
            _min = np.nanmin(self.elevation(row_range=slice(start, stop)))
            if np.isnan(elv_min):
                elv_min = _min
            else:
                if _min < elv_min:
                    elv_min = _min
            _max = np.nanmax(self.elevation(row_range=slice(start, stop)))
            if np.isnan(elv_max):
                elv_max = _max
            else:
                if _max > elv_max:
                    elv_max = _max

        return elv_min, elv_max

    def depth_min_max(self) -> Tuple[float, float]:
        elv_min, elv_max = self.elevation_min_max()
        return -elv_max, -elv_min

    def vr_refinements_shape(self):
        return self[BAGFile._bag_varres_refinements].shape

    def vr_elevation_min_max(self) -> Tuple[float, float]:
        # rows, cols = self.vr_refinements_shape()
        # logger.debug('refinements shape: %s, %s' % (rows, cols))

        vr_el = self[BAGFile._bag_varres_refinements][0]['depth']
        mask = vr_el == BAGFile.BAG_NAN
        vr_el[mask] = np.nan

        # logger.debug(vr_el)

        return np.nanmin(vr_el), np.nanmax(vr_el)

    def vr_depth_min_max(self) -> Tuple[float, float]:
        elv_min, elv_max = self.vr_elevation_min_max()
        return -elv_max, -elv_min

    def has_uncertainty(self):
        return BAGFile._bag_uncertainty in self

    def has_product_uncertainty(self):
        if self.has_uncertainty() and \
                (self.meta.unc_type == "productUncert" or self.meta.unc_type == "ProductUncert"):  # Leidos bug
            return True
        return False

    def uncertainty(self, mask_nan=True, row_range=None):
        """
        Return the uncertainty as numpy array

        mask_nan
            If True, apply a mask using the BAG nan value
        row_range
            If present, a slice of rows to read from
        """
        if row_range:
            if not isinstance(row_range, slice):
                raise BAGError("Invalid type of slice selector: %s" % type(row_range))
            if (row_range.start < 0) or (row_range.start >= self.uncertainty_shape()[0]) \
                    or (row_range.stop < 0) or (row_range.stop > self.uncertainty_shape()[0]) \
                    or (row_range.start > row_range.stop):
                raise BAGError("Invalid values for slice selector: %s" % row_range)

        if mask_nan:
            if row_range:
                un = self[BAGFile._bag_uncertainty][row_range]
            else:
                un = self[BAGFile._bag_uncertainty][:]
            mask = un == BAGFile.BAG_NAN
            un[mask] = np.nan
            return un

        if row_range:
            return self[BAGFile._bag_uncertainty][row_range]
        else:
            return self[BAGFile._bag_uncertainty][:]

    def uncertainty_shape(self):
        return self[BAGFile._bag_uncertainty].shape

    def uncertainty_min_max(self) -> Tuple[float, float]:
        rows, cols = self.uncertainty_shape()
        # logger.debug('shape: %s, %s' % (rows, cols))

        mem_row = cols * 32 / 1024 / 1024
        # mem = mem_row * rows
        # logger.debug('estimated memory: %.1f MB' % mem)
        chunk_size = 8096
        chunk_rows = int(chunk_size / mem_row) + 1
        # logger.debug('nr of rows per chunk: %s' % chunk_rows)

        unc_min = np.nan
        unc_max = np.nan
        for start in range(0, rows, chunk_rows):
            stop = start + chunk_rows
            if stop > rows:
                stop = rows
            # logger.debug('slice: %s-%s' % (start, stop))
            _min = np.nanmin(self.uncertainty(row_range=slice(start, stop)))
            if np.isnan(unc_min):
                unc_min = _min
            else:
                if _min < unc_min:
                    unc_min = _min
            _max = np.nanmax(self.uncertainty(row_range=slice(start, stop)))
            if np.isnan(unc_max):
                unc_max = _max
            else:
                if _max > unc_max:
                    unc_max = _max

        return unc_min, unc_max

    def uncertainty_greater_than(self, th: float) -> list[list[int | float]]:
        rows, cols = self.uncertainty_shape()
        # logger.debug('shape: %s, %s' % (rows, cols))

        self.populate_metadata()

        x_min = self.meta.sw[0]
        y_min = self.meta.sw[1]
        x_res = self.meta.res_x
        y_res = self.meta.res_y
        # logger.debug("info: %f %f %f %f" % (x_min, y_min, x_res, y_res))

        in_srs = osr.SpatialReference()
        in_srs.ImportFromWkt(self.meta.wkt_srs)
        if in_srs.IsCompound():
            in_srs.StripVertical()
        out_srs = osr.SpatialReference()
        out_srs.ImportFromEPSG(4326)
        out_srs.SetAxisMappingStrategy(osr.OAMS_TRADITIONAL_GIS_ORDER)
        ctr = osr.CoordinateTransformation(in_srs, out_srs)

        mem_row = cols * 32 / 1024 / 1024
        # mem = mem_row * rows
        # logger.debug('estimated memory: %.1f MB' % mem)
        chunk_size = 8096
        chunk_rows = int(chunk_size / mem_row) + 1
        # logger.debug('nr of rows per chunk: %s' % chunk_rows)

        xyz = list()
        for start in range(0, rows, chunk_rows):
            stop = start + chunk_rows
            if stop > rows:
                stop = rows

            unc = self.uncertainty(row_range=slice(start, stop))
            ijs = np.argwhere(unc > th)
            for ij in ijs:
                i = ij[0]
                j = ij[1]
                e = x_min + j * x_res
                n = y_min + (start + i) * y_res
                lat, lon, _ = ctr.TransformPoint(e, n)
                u = float(unc[i, j])
                xyz.append([float(lat), float(lon), u])
                # logger.info("%d,%d: %.7f %.7f %.3f" % ((start + i), j, xyz[-1][0], xyz[-1][1], xyz[-1][2]))

        return xyz

    def uncertainty_has_depth(self) -> list[list[int | float]]:
        rows, cols = self.uncertainty_shape()
        # logger.debug('shape: %s, %s (%d)' % (rows, cols, rows * cols))

        self.populate_metadata()

        x_min = self.meta.sw[0]
        y_min = self.meta.sw[1]
        x_res = self.meta.res_x
        y_res = self.meta.res_y
        # logger.debug("info: %f %f %f %f" % (x_min, y_min, x_res, y_res))

        in_srs = osr.SpatialReference()
        in_srs.ImportFromWkt(self.meta.wkt_srs)
        if in_srs.IsCompound():
            in_srs.StripVertical()
        out_srs = osr.SpatialReference()
        out_srs.ImportFromEPSG(4326)
        out_srs.SetAxisMappingStrategy(osr.OAMS_TRADITIONAL_GIS_ORDER)
        ctr = osr.CoordinateTransformation(in_srs, out_srs)

        mem_row = cols * 32 / 1024 / 1024
        # mem = mem_row * rows
        # logger.debug('estimated memory: %.1f MB' % mem)
        chunk_size = 8096
        chunk_rows = int(chunk_size / mem_row) + 1
        # logger.debug('nr of rows per chunk: %s' % chunk_rows)

        xyz = list()
        for start in range(0, rows, chunk_rows):
            stop = start + chunk_rows
            if stop > rows:
                stop = rows

            unc = self.uncertainty(row_range=slice(start, stop))
            # logger.info("unc: %d, %d" % (np.isnan(unc).sum(), np.isfinite(unc).sum()))
            dep = self.elevation(row_range=slice(start, stop))
            # logger.info("dep: %d, %d" % (np.isnan(dep).sum(), np.isfinite(dep).sum()))
            unc[np.isfinite(dep)] = np.nan
            # logger.info(unc)
            ijs = np.argwhere(np.isfinite(unc))
            # logger.info(ijs)
            for ij in ijs:
                i = ij[0]
                j = ij[1]
                e = x_min + j * x_res
                n = y_min + (start + i) * y_res
                lat, lon, _ = ctr.TransformPoint(e, n)
                u = float(unc[i, j])
                xyz.append([float(lat), float(lon), u])
                # logger.info("%d,%d: %.7f %.7f %.3f" % ((start + i), j, xyz[-1][0], xyz[-1][1], xyz[-1][2]))

        return xyz

    def depth_has_uncertainty(self) -> list[list[int | float]]:
        rows, cols = self.uncertainty_shape()
        # logger.debug('shape: %s, %s' % (rows, cols))

        self.populate_metadata()

        x_min = self.meta.sw[0]
        y_min = self.meta.sw[1]
        x_res = self.meta.res_x
        y_res = self.meta.res_y
        # logger.debug("info: %f %f %f %f" % (x_min, y_min, x_res, y_res))

        in_srs = osr.SpatialReference()
        in_srs.ImportFromWkt(self.meta.wkt_srs)
        if in_srs.IsCompound():
            in_srs.StripVertical()
        out_srs = osr.SpatialReference()
        out_srs.ImportFromEPSG(4326)
        out_srs.SetAxisMappingStrategy(osr.OAMS_TRADITIONAL_GIS_ORDER)
        ctr = osr.CoordinateTransformation(in_srs, out_srs)

        mem_row = cols * 32 / 1024 / 1024
        # mem = mem_row * rows
        # logger.debug('estimated memory: %.1f MB' % mem)
        chunk_size = 8096
        chunk_rows = int(chunk_size / mem_row) + 1
        # logger.debug('nr of rows per chunk: %s' % chunk_rows)

        xyz = list()
        for start in range(0, rows, chunk_rows):
            stop = start + chunk_rows
            if stop > rows:
                stop = rows

            unc = self.uncertainty(row_range=slice(start, stop))
            # logger.info(unc)
            dep = self.elevation(row_range=slice(start, stop))
            # logger.info(dep)
            dep[np.isfinite(unc)] = np.nan
            # logger.info(dep)
            ijs = np.argwhere(np.isfinite(dep))
            # logger.info(ijs)
            for ij in ijs:
                i = ij[0]
                j = ij[1]
                e = x_min + j * x_res
                n = y_min + (start + i) * y_res
                lat, lon, _ = ctr.TransformPoint(e, n)
                z = -float(dep[i, j])
                xyz.append([float(lat), float(lon), z])
                # logger.info("%d,%d: %.7f %.7f %.3f" % ((start + i), j, xyz[-1][0], xyz[-1][1], xyz[-1][2]))

        return xyz

    def vr_uncertainty_min_max(self) -> Tuple[float, float]:
        # rows, cols = self.vr_refinements_shape()
        # logger.debug('shape: %s, %s' % (rows, cols))

        vr_unc = self[BAGFile._bag_varres_refinements][0]['depth_uncrt']
        mask = vr_unc == BAGFile.BAG_NAN
        vr_unc[mask] = np.nan

        # logger.debug(vr_el)

        return np.nanmin(vr_unc), np.nanmax(vr_unc)

    def vr_uncertainty_greater_than(self, th: float) -> list[list[int | float]]:
        # rows, cols = self.vr_refinements_shape()
        # logger.debug('shape: %s, %s' % (rows, cols))

        self.populate_metadata()

        x_min = self.meta.sw[0]
        y_min = self.meta.sw[1]
        x_res = self.meta.res_x
        y_res = self.meta.res_y
        # logger.debug("info: %f %f %f %f" % (x_min, y_min, x_res, y_res))

        in_srs = osr.SpatialReference()
        in_srs.ImportFromWkt(self.meta.wkt_srs)
        if in_srs.IsCompound():
            in_srs.StripVertical()
        out_srs = osr.SpatialReference()
        out_srs.ImportFromEPSG(4326)
        out_srs.SetAxisMappingStrategy(osr.OAMS_TRADITIONAL_GIS_ORDER)
        ctr = osr.CoordinateTransformation(in_srs, out_srs)

        vr_unc = self[BAGFile._bag_varres_refinements][0]['depth_uncrt']
        mask = vr_unc == BAGFile.BAG_NAN
        vr_unc[mask] = np.nan

        xyz_dict = dict()
        for idx, unc in enumerate(vr_unc):
            if unc > th:
                xyz_dict[idx] = unc

        # logger.info("Located %d outliers" % len(xyz_dict))

        xyz = list()
        vr_ixs = self[BAGFile._bag_varres_metadata][:]
        rows, cols = vr_ixs.shape
        i = 0
        for sg_r in range(rows):
            for sg_c in range(cols):
                if vr_ixs[sg_r, sg_c][1] == 0:
                    continue
                ir = vr_ixs[sg_r, sg_c][1] * vr_ixs[sg_r, sg_c][2]
                for ir_idx in range(ir):
                    j = i + ir_idx
                    if j not in xyz_dict:
                        continue
                    unc = float(xyz_dict[j])
                    # logger.debug("Located outliers: %d %f in %d,%d: %s" % (j, unc, sg_r, sg_c, vr_ixs[sg_r, sg_c]))
                    # vr_ixs[r, c]
                    rfn_r = ir_idx // vr_ixs[sg_r, sg_c][1]
                    rfn_c = ir_idx % vr_ixs[sg_r, sg_c][1]
                    # logger.debug("%d > %d,%d" % (ir_idx, rfn_r, rfn_c))
                    e = x_min + (sg_c - 0.5) * x_res + vr_ixs[sg_r, sg_c][5] + rfn_c * vr_ixs[sg_r, sg_c][3]
                    n = y_min + (sg_r - 0.5) * y_res + vr_ixs[sg_r, sg_c][6] + rfn_r * vr_ixs[sg_r, sg_c][4]
                    lat, lon, _ = ctr.TransformPoint(e, n)
                    xyz.append([float(lat), float(lon), unc])
                i += ir

        return xyz

    def vr_depth_has_uncertainty(self) -> list[list[int | float]]:
        # rows, cols = self.vr_refinements_shape()
        # logger.debug('shape: %s, %s' % (rows, cols))

        self.populate_metadata()

        x_min = self.meta.sw[0]
        y_min = self.meta.sw[1]
        x_res = self.meta.res_x
        y_res = self.meta.res_y
        # logger.debug("info: %f %f %f %f" % (x_min, y_min, x_res, y_res))

        in_srs = osr.SpatialReference()
        in_srs.ImportFromWkt(self.meta.wkt_srs)
        if in_srs.IsCompound():
            in_srs.StripVertical()
        out_srs = osr.SpatialReference()
        out_srs.ImportFromEPSG(4326)
        out_srs.SetAxisMappingStrategy(osr.OAMS_TRADITIONAL_GIS_ORDER)
        ctr = osr.CoordinateTransformation(in_srs, out_srs)

        vr_unc = self[BAGFile._bag_varres_refinements][0]['depth_uncrt']
        mask = vr_unc == BAGFile.BAG_NAN
        vr_unc[mask] = np.nan
        vr_dep = self[BAGFile._bag_varres_refinements][0]['depth']
        mask = vr_dep == BAGFile.BAG_NAN
        vr_dep[mask] = np.nan

        xyz_dict = dict()
        for idx, unc in enumerate(vr_unc):
            dep = vr_dep[idx]
            if np.isfinite(dep) and np.isnan(unc):
                xyz_dict[idx] = dep

        # logger.info("Located %d outliers" % len(xyz_dict))

        xyz = list()
        vr_ixs = self[BAGFile._bag_varres_metadata][:]
        rows, cols = vr_ixs.shape
        i = 0
        for sg_r in range(rows):
            for sg_c in range(cols):
                if vr_ixs[sg_r, sg_c][1] == 0:
                    continue
                ir = vr_ixs[sg_r, sg_c][1] * vr_ixs[sg_r, sg_c][2]
                for ir_idx in range(ir):
                    j = i + ir_idx
                    if j not in xyz_dict:
                        continue
                    dep = float(xyz_dict[j])
                    # logger.debug("Located outliers: %d %f in %d,%d: %s" % (j, unc, sg_r, sg_c, vr_ixs[sg_r, sg_c]))
                    # vr_ixs[r, c]
                    rfn_r = ir_idx // vr_ixs[sg_r, sg_c][1]
                    rfn_c = ir_idx % vr_ixs[sg_r, sg_c][1]
                    # logger.debug("%d > %d,%d" % (ir_idx, rfn_r, rfn_c))
                    e = x_min + (sg_c - 0.5) * x_res + vr_ixs[sg_r, sg_c][5] + rfn_c * vr_ixs[sg_r, sg_c][3]
                    n = y_min + (sg_r - 0.5) * y_res + vr_ixs[sg_r, sg_c][6] + rfn_r * vr_ixs[sg_r, sg_c][4]
                    lat, lon, _ = ctr.TransformPoint(e, n)
                    xyz.append([float(lat), float(lon), dep])
                i += ir

        return xyz

    def vr_uncertainty_has_depth(self) -> list[list[int | float]]:
        # rows, cols = self.vr_refinements_shape()
        # logger.debug('shape: %s, %s' % (rows, cols))

        self.populate_metadata()

        x_min = self.meta.sw[0]
        y_min = self.meta.sw[1]
        x_res = self.meta.res_x
        y_res = self.meta.res_y
        # logger.debug("info: %f %f %f %f" % (x_min, y_min, x_res, y_res))

        in_srs = osr.SpatialReference()
        in_srs.ImportFromWkt(self.meta.wkt_srs)
        if in_srs.IsCompound():
            in_srs.StripVertical()
        out_srs = osr.SpatialReference()
        out_srs.ImportFromEPSG(4326)
        out_srs.SetAxisMappingStrategy(osr.OAMS_TRADITIONAL_GIS_ORDER)
        ctr = osr.CoordinateTransformation(in_srs, out_srs)

        vr_unc = self[BAGFile._bag_varres_refinements][0]['depth_uncrt']
        mask = vr_unc == BAGFile.BAG_NAN
        vr_unc[mask] = np.nan
        vr_dep = self[BAGFile._bag_varres_refinements][0]['depth']
        mask = vr_dep == BAGFile.BAG_NAN
        vr_dep[mask] = np.nan

        xyz_dict = dict()
        for idx, unc in enumerate(vr_unc):
            dep = vr_dep[idx]
            if np.isfinite(unc) and np.isnan(dep):
                xyz_dict[idx] = unc

        # logger.info("Located %d outliers" % len(xyz_dict))

        xyz = list()
        vr_ixs = self[BAGFile._bag_varres_metadata][:]
        rows, cols = vr_ixs.shape
        i = 0
        for sg_r in range(rows):
            for sg_c in range(cols):
                if vr_ixs[sg_r, sg_c][1] == 0:
                    continue
                ir = vr_ixs[sg_r, sg_c][1] * vr_ixs[sg_r, sg_c][2]
                for ir_idx in range(ir):
                    j = i + ir_idx
                    if j not in xyz_dict:
                        continue
                    unc = float(xyz_dict[j])
                    # logger.debug("Located outliers: %d %f in %d,%d: %s" % (j, unc, sg_r, sg_c, vr_ixs[sg_r, sg_c]))
                    # vr_ixs[r, c]
                    rfn_r = ir_idx // vr_ixs[sg_r, sg_c][1]
                    rfn_c = ir_idx % vr_ixs[sg_r, sg_c][1]
                    # logger.debug("%d > %d,%d" % (ir_idx, rfn_r, rfn_c))
                    e = x_min + (sg_c - 0.5) * x_res + vr_ixs[sg_r, sg_c][5] + rfn_c * vr_ixs[sg_r, sg_c][3]
                    n = y_min + (sg_r - 0.5) * y_res + vr_ixs[sg_r, sg_c][6] + rfn_r * vr_ixs[sg_r, sg_c][4]
                    lat, lon, _ = ctr.TransformPoint(e, n)
                    xyz.append([float(lat), float(lon), unc])
                i += ir

        return xyz

    def has_density(self):
        # noinspection PyBroadException
        try:
            self[BAGFile._bag_elevation_solution]['num_soundings']
        except Exception:
            return False
        return True

    def density(self, mask_nan=True, row_range=None):
        """
        Return the density as numpy array

        mask_nan
            If True, apply a mask using the BAG nan value
        row_range
            If present, a slice of rows to read from
        """
        if row_range:
            if not isinstance(row_range, slice):
                raise BAGError("Invalid type of slice selector: %s" % type(row_range))
            if (row_range.start < 0) or (row_range.start >= self.density_shape()[0]) \
                    or (row_range.stop < 0) or (row_range.stop > self.density_shape()[0]) \
                    or (row_range.start > row_range.stop):
                raise BAGError("Invalid values for slice selector: %s" % row_range)

        if mask_nan:
            if row_range:
                de = self[BAGFile._bag_elevation_solution]['num_soundings'][row_range]
            else:
                de = self[BAGFile._bag_elevation_solution]['num_soundings'][:]
            de = de.astype(float)
            mask = de == BAGFile.BAG_NAN
            de[mask] = np.nan
            return de

        if row_range:
            de = self[BAGFile._bag_elevation_solution]['num_soundings'][row_range]
        else:
            de = self[BAGFile._bag_elevation_solution]['num_soundings'][:]
        de = de.astype(float)
        return de

    def density_shape(self):
        return self[BAGFile._bag_elevation_solution].shape

    def has_tracking_list(self):
        return BAGFile._bag_tracking_list in self

    def tracking_list(self):
        """ Return the tracking list as numpy array """
        return self[BAGFile._bag_tracking_list][:]

    def tracking_list_fields(self):
        """ Return the tracking list field names """
        return self[BAGFile._bag_tracking_list].dtype.names

    def tracking_list_types(self):
        """ Return the tracking list field names """
        return self[BAGFile._bag_tracking_list].dtype

    def has_valid_row_in_tracking_list(self):
        rows, _ = self.elevation_shape()
        # logger.info('rows: %s, cols: %s' % (rows, _))

        tl = self.tracking_list()
        for idx, row in enumerate(tl['row']):
            if row >= rows:
                logger.warning("%d 'row' entry is invalid: %s" % (idx, row))
                return False

        return True

    def has_valid_col_in_tracking_list(self):
        rows, cols = self.elevation_shape()
        # logger.info('rows: %s, cols: %s' % (rows, cols))

        tl = self.tracking_list()
        for idx, col in enumerate(tl['col']):
            if col >= cols:
                logger.warning("%d 'col' entry is invalid: %s" % (idx, col))
                return False

        return True

    def metadata(self, as_string: bool = True, as_pretty_xml: bool = True) -> bytes | str:
        """ Return the metadata

        as_string
            If True, convert the metadata from a dataset of characters to a string
        as_pretty_xml
            If True, return the xml in a pretty format as bytes
        """
        xml_bytes = self[BAGFile._bag_metadata][:].tostring().strip(b'\x00')

        if as_pretty_xml:
            xml_tree = etree.fromstring(xml_bytes)
            pretty_bytes = etree.tostring(xml_tree, pretty_print=True)
            if as_string:
                return pretty_bytes.decode()
            return pretty_bytes
        else:
            if as_string:
                return xml_bytes.decode()
            return xml_bytes

    def extract_metadata(self, name=None):
        """ Save metadata on disk

        name
            The file path where the metadata will be saved. If None, use a default name.
        """

        meta_xml = self.metadata(as_pretty_xml=True)
        if meta_xml is None:
            logger.info("unable to access the metadata")
            return

        if name is None:
            name = os.path.join(self.default_metadata_file)

        with open(os.path.abspath(name), 'w') as fid:
            fid.write(meta_xml.decode())

    def substitute_metadata(self, path):
        """ Substitute internal metadata

        name
            The file path where the new metadata are.
        """

        path = os.path.abspath(path)
        if not os.path.exists(path):
            logger.info("the passed file does not exist")
            return

        with open(path, 'r') as fid:
            xml_string = str.encode(fid.read())

        is_valid = self.validate_metadata(xml_string)
        if not is_valid:
            logger.info("the passed metadata file is not valid")
            return

        del self[BAGFile._bag_metadata]
        xml_sz = len(xml_string)
        ds = self.create_dataset(self._bag_metadata, (xml_sz,), dtype="S1")
        for i, bt in enumerate(xml_string):
            ds[i] = bytes([bt])

    def validate_metadata(self, xml_string: None | bytes = None) -> bool:
        """ Validate metadata based on XML Schemas and schematron. """
        # clean metadata error list
        self.meta_errors = list()
        # assuming a valid BAG
        is_valid = True

        if xml_string is None:
            xml_string = self.metadata(as_pretty_xml=True)

        try:
            xml_tree = etree.fromstring(xml_string)
        except etree.Error as e:
            logger.warning("unable to parse XML metadata: %s" % e)
            self.meta_errors.append(e)
            return False

        try:
            schema_path = os.path.join(Helper.iso19139_folder(), 'bag', 'bag.xsd')
            schema_doc = etree.parse(schema_path)
            schema = etree.XMLSchema(schema_doc)
        except etree.Error as e:
            logger.warning("unable to parse XML schema: %s" % e)
            self.meta_errors.append(e)
            return False

        try:
            schema.assertValid(xml_tree)
        except etree.DocumentInvalid as e:
            logger.warning("invalid metadata based on XML schema: %s" % e)
            self.meta_errors.append(e)
            for i in schema.error_log:
                self.meta_errors.append(i)
            is_valid = False

        if is_valid:
            logger.debug("xsd validated")

        try:
            schematron_path = os.path.join(Helper.iso19757_3_folder(), 'bag_metadata_profile.sch')
            schematron_doc = etree.parse(schematron_path)
        except etree.DocumentInvalid as e:
            logger.warning("unable to parse BAG schematron: %s" % e)
            self.meta_errors.append(e)
            return False

        try:
            schematron = isoschematron.Schematron(schematron_doc, store_report=True)
        except etree.DocumentInvalid as e:
            logger.warning("unable to load BAG schematron: %s" % e)
            self.meta_errors.append(e)
            return False

        if schematron.validate(xml_tree):
            logger.debug("schematron validated")
        else:
            logger.warning("invalid metadata based on Schematron")
            is_valid = False
            ns = {
                'svrl': 'http://purl.oclc.org/dsdl/svrl',
            }
            for i in schematron.error_log:
                err_tree = etree.fromstring(i.message)
                # print(etree.tostring(err_tree, pretty_print=True))
                err_msg = err_tree.xpath('/svrl:failed-assert/svrl:text', namespaces=ns)[0].text.strip()
                logger.warning(err_msg)
                self.meta_errors.append(err_msg)

        return is_valid

    def validation_info(self):
        """ Return a message string with the result of the validation """
        msg = str()

        msg += "XML input source: %s\nValidation output: " % self._bag_metadata
        if self.validate_metadata():
            msg += "VALID"
        else:
            msg += "INVALID\nReasons:\n"
            for err_msg in self.meta_errors:
                msg += " - %s\n" % err_msg
        return msg

    def populate_metadata(self):
        """ Populate metadata class """

        if self.meta is not None:
            # log.debug("metadata already populated")
            return self.meta

        self.meta = Meta(meta_xml=self.metadata(as_pretty_xml=True))
        return self.meta

    def modify_wkt_prj(self, wkt_hor, wkt_ver=None):
        """ Modify the wkt prj in the metadata content

        text
            The new wkt prj text to use
        """
        ns = {
            'bag': 'http://www.opennavsurf.org/schema/bag',
            'gco': 'http://www.isotc211.org/2005/gco',
            'gmd': 'http://www.isotc211.org/2005/gmd',
            'gmi': 'http://www.isotc211.org/2005/gmi',
            'gml': 'http://www.opengis.net/gml/3.2',
            'xsi': 'http://www.w3.org/2001/XMLSchema-instance',
        }

        # print(self[BAGFile._bag_metadata][:])
        xml_tree = etree.fromstring(self[BAGFile._bag_metadata][:].tostring())

        try:
            ret = xml_tree.xpath('//*/gmd:referenceSystemInfo/gmd:MD_ReferenceSystem/'
                                 'gmd:referenceSystemIdentifier/gmd:RS_Identifier/gmd:code/gco:CharacterString',
                                 namespaces=ns)
            ret[0].text = wkt_hor
            if wkt_ver is not None:
                ret[1].text = wkt_ver

        except etree.Error as e:
            logger.warning("unable to read the WKT projection string: %s" % e)
            return

        new_xml = etree.tostring(xml_tree, pretty_print=True)
        del self[BAGFile._bag_metadata]
        ds = self.create_dataset(BAGFile._bag_metadata, shape=(len(new_xml),), dtype="S1")
        for i, x in enumerate(new_xml):
            ds[i] = bytes([x])

    def modify_bbox(self, west, east, south, north):
        """ attempts to modify the bounding box values """
        ns = {
            'bag': 'http://www.opennavsurf.org/schema/bag',
            'gco': 'http://www.isotc211.org/2005/gco',
            'gmd': 'http://www.isotc211.org/2005/gmd',
            'gmi': 'http://www.isotc211.org/2005/gmi',
            'gml': 'http://www.opengis.net/gml/3.2',
            'xsi': 'http://www.w3.org/2001/XMLSchema-instance',
        }

        xml_tree = etree.fromstring(self[BAGFile._bag_metadata][:].tostring())

        try:
            ret_x_min = xml_tree.xpath('//*/gmd:EX_GeographicBoundingBox/gmd:westBoundLongitude/gco:Decimal',
                                       namespaces=ns)
            ret_x_max = xml_tree.xpath('//*/gmd:EX_GeographicBoundingBox/gmd:eastBoundLongitude/gco:Decimal',
                                       namespaces=ns)
        except etree.Error as e:
            logger.warning("unable to read the bbox's longitude values: %s" % e)
            return

        try:
            ret_x_min[0].text = "%s" % west
            ret_x_max[0].text = "%s" % east
        except (ValueError, IndexError) as e:
            logger.warning("unable to read the bbox's longitude values: %s" % e)
            return

        try:
            ret_y_min = xml_tree.xpath('//*/gmd:EX_GeographicBoundingBox/gmd:southBoundLatitude/gco:Decimal',
                                       namespaces=ns)
            ret_y_max = xml_tree.xpath('//*/gmd:EX_GeographicBoundingBox/gmd:northBoundLatitude/gco:Decimal',
                                       namespaces=ns)
        except etree.Error as e:
            logger.warning("unable to read the bbox's latitude values: %s" % e)
            return

        try:
            ret_y_min[0].text = "%s" % south
            ret_y_max[0].text = "%s" % north
        except (ValueError, IndexError) as e:
            logger.warning("unable to read the bbox's latitude values: %s" % e)
            return

        new_xml = etree.tostring(xml_tree, pretty_print=True)
        del self[BAGFile._bag_metadata]
        ds = self.create_dataset(BAGFile._bag_metadata, shape=(len(new_xml),), dtype="S1")
        for i, x in enumerate(new_xml):
            ds[i] = bytes([x])

    def has_varres_metadata(self):
        return BAGFile._bag_varres_metadata in self

    def has_varres_refinements(self):
        return BAGFile._bag_varres_refinements in self

    def has_varres_tracking_list(self):
        return BAGFile._bag_varres_tracking_list in self

    def _str_group_info(self, grp):
        if grp == self._bag_root:
            self._str += "  <root>\n"
        elif grp == self._bag_elevation:
            self._str += "  <elevation shape=%s>\n" % str(self.elevation().shape)
        elif grp == self._bag_uncertainty:
            self._str += "  <uncertainty shape=%s>\n" % str(self.uncertainty().shape)
        elif grp == self._bag_tracking_list:
            self._str += "  <tracking list shape=%s>\n" % str(self.tracking_list().shape)
        elif grp == self._bag_metadata:
            if self.meta is not None:
                self._str += "  %s\n" % str(self.meta)
            else:
                self._str += "  <%s>\n" % grp
        else:
            self._str += "  <%s>\n" % grp

        if grp != self._bag_metadata:
            for atr in self[grp].attrs:
                atr_val = self[grp].attrs[atr]
                self._str += "    <%s: %s (%s, %s)>\n" % (atr, atr_val, atr_val.shape, atr_val.dtype)

    def __str__(self):
        self._str = super(BAGFile, self).__str__()
        self.visit(self._str_group_info)
        return self._str



import os
import logging
import argparse

from hyo2.abc2.lib.logging import set_logging
from hyo2.bag import __version__
from hyo2.bag.bag import BAGFile, is_bag

logger = logging.getLogger(__name__)


def main():
    app_name = "bag_validate"
    app_info = "Validation of an OpenNS BAG file, using hyo2.bag r%s" % __version__

    parser = argparse.ArgumentParser(prog=app_name, description=app_info)
    parser.add_argument("bag_file", type=str, help="a valid BAG file to validate")
    parser.add_argument("-v", "--verbose", help="increase output verbosity", action="store_true")
    args = parser.parse_args()

    if args.verbose:
        set_logging(ns_list=['hyo2.bag'])
        logger.debug("> verbosity: ON")

        logger.debug("> input: %s" % args.bag_file)

    if not os.path.exists(args.bag_file):
        parser.exit(1, "ERROR: the input valid does not exist: %s" % args.bag_file)

    if not is_bag(args.bag_file):
        parser.exit(1, "ERROR: the input valid does not seem a BAG file: %s" % args.bag_file)

    bf = BAGFile(args.bag_file, mode='r')
    logger.debug(bf.validation_info())


if __name__ == "__main__":
    main()



import os
import logging
from matplotlib import pyplot as plt

from hyo2.abc2.lib.logging import set_logging
from hyo2.bag.bag import BAGFile
from hyo2.bag.bag import BAGError
from hyo2.bag.helper import Helper
from hyo2.bag.tracklist import TrackList2Csv

logger = logging.getLogger(__name__)
set_logging(ns_list=['hyo2.bag'])

bag_file = os.path.join(Helper.samples_folder(), "bdb_01.bag")
bag = BAGFile(bag_file, mode='r')
tl = bag.tracking_list()
tl2csv = TrackList2Csv(tl)

bag_file = os.path.join(Helper.samples_folder(), "bdb_02.bag")
bag = BAGFile(bag_file, mode='r')
tl = bag.tracking_list()
tl_fields = bag.tracking_list_fields()
tl_types = bag.tracking_list_types()
logger.debug(tl_fields)
logger.debug(tl_types)
logger.debug(tl)
_ = TrackList2Csv(track_list=tl, header=tl_fields)



import unittest
import os

from hyo2.bag.helper import Helper
from hyo2.bag.base import is_bag, File


class TestBagBase(unittest.TestCase):

    def setUp(self):
        self.file_bag_0 = os.path.join(Helper.samples_folder(), "bdb_01.bag")
        self.file_bag_1 = os.path.join(Helper.samples_folder(), "bdb_02.bag")
        self.file_fake_0 = os.path.join(Helper.samples_folder(), "fake_00.bag")

    def tearDown(self):
        pass

    def test_is_bag(self):
        self.assertTrue(is_bag(self.file_bag_0))
        self.assertTrue(is_bag(self.file_bag_1))
        self.assertFalse(is_bag(self.file_fake_0))

    def test_bag_File_raise(self):
        with self.assertRaises(IOError):
            File(self.file_fake_0)

    def test_bag_File_open(self):
        self.assertIsNotNone(File(self.file_bag_0))
        self.assertIsNotNone(File(self.file_bag_1))

    def test_bag_File_filename(self):
        bag_0 = File(self.file_bag_0)
        self.assertEqual(os.path.abspath(self.file_bag_0), bag_0.filename)
        bag_1 = File(self.file_bag_1)
        self.assertEqual(os.path.abspath(self.file_bag_1), bag_1.filename)


def suite():
    s = unittest.TestSuite()
    s.addTests(unittest.TestLoader().loadTestsFromTestCase(TestBagBase))
    return s



import os
import logging

from hyo2.abc2.lib.logging import set_logging
from hyo2.bag.helper import BAGError, Helper

logger = logging.getLogger(__name__)
set_logging(ns_list=['hyo2.bag'])

try:
    raise BAGError("test")
except BAGError as e:
    logger.debug(e)

data_folder = Helper.samples_folder()
if os.path.exists(data_folder):
    logger.debug("data folder: %s" % data_folder)

iso_folder = Helper.iso19139_folder()
if os.path.exists(iso_folder):
    logger.debug("iso folder: %s" % iso_folder)

iso_folder = Helper.iso19757_3_folder()
if os.path.exists(iso_folder):
    logger.debug("iso folder: %s" % iso_folder)



import os
import logging
from osgeo import osr

from hyo2.abc2.lib.logging import set_logging
from hyo2.bag.bag import BAGFile
from hyo2.bag.helper import Helper
from hyo2.bag.meta import Meta

set_logging(ns_list=['hyo2.bag'])
logger = logging.getLogger(__name__)

# file_bag_0 = os.path.join(Helper.samples_folder(), "bdb_01.bag")
file_bag_0 = r"C:\Users\gmasetti\Downloads\H13745_MB_4m_MLLW_1of1 (1).bag"
if os.path.exists(file_bag_0):
    logger.debug("- file_bag_0: %s" % file_bag_0)

bag_0 = BAGFile(file_bag_0, mode='r')
logger.debug(bag_0)

meta = Meta(bag_0.metadata())

logger.debug("vertical datum: %s (%s)" % (meta.wkt_vertical_datum, type(meta.wkt_vertical_datum)))

vrs = osr.SpatialReference()
ret = vrs.ImportFromWkt(meta.wkt_vertical_datum)
logger.debug("wkt import: %s" % ret)

is_vertical = vrs.IsVertical() == 1
logger.debug('wkt is vertical: %s' % is_vertical)

is_epsg = vrs.GetAttrValue("AUTHORITY", 0).lower() == 'epsg'
if is_epsg:
    epsg = vrs.GetAttrValue("AUTHORITY", 1)
    logger.debug("EPSG:%s" % epsg)

is_depth = vrs.GetAttrValue("AXIS", 0).lower() == 'depth'
if is_depth:
    direction = vrs.GetAttrValue("AXIS", 1)
    logger.debug("depth is %s" % direction)



import logging
import subprocess
import os

from hyo2.abc2.lib.logging import set_logging
from hyo2.bag.helper import Helper
from hyo2.bag import tools

logger = logging.getLogger(__name__)
set_logging(ns_list=['hyo2.bag'])

tools_folder = os.path.abspath(os.path.dirname(tools.__file__))
tool_path = os.path.join(tools_folder, 'bag_uncertainty.py')
logger.debug("tool: %s" % tool_path)

# help
logger.debug("# -h")
subprocess.call("python %s -h" % tool_path)

# verbose + test file
file_bag_0 = os.path.join(Helper.samples_folder(), "bdb_01.bag")
logger.debug("# -v %s" % file_bag_0)
subprocess.call("python %s -v -o test_unc.tiff %s" % (tool_path, file_bag_0))
subprocess.call("python %s -v -o test_unc.ascii -f ascii %s" % (tool_path, file_bag_0))

# verbose + fake file
file_bag_0 = os.path.join(Helper.samples_folder(), "not_present_00.bag")
logger.debug("# -v %s" % file_bag_0)
subprocess.call("python %s -v %s" % (tool_path, file_bag_0))

# test file
file_bag_0 = os.path.join(Helper.samples_folder(), "bdb_01.bag")
logger.debug("# %s" % file_bag_0)
subprocess.call("python %s %s" % (tool_path, file_bag_0))



from h5py import File
from lxml import etree, __version__ as lxml_version

bag_path = r"C:\code\cpp\BAG\examples\sample-data\sample-1.5.0.bag"

strip_x00 = True

print("lxml version: %s" % lxml_version)
print("libxml version: %s" % (etree.LIBXML_COMPILED_VERSION, ))

bag = File(bag_path, 'r')
xml = bag["BAG_root/metadata"][:].tobytes()
if strip_x00:
    xml = xml.strip(b'\x00')
xml_tree = etree.fromstring(xml)



import unittest
import os

from hyo2.bag.helper import Helper
from hyo2.bag.helper import BAGError


class TestBAGError(unittest.TestCase):

    def setUp(self):
        self.err = BAGError("test")

    def tearDown(self):
        pass

    def test_is_instance(self):
        self.assertTrue(isinstance(self.err, Exception))

    def test_raise(self):
        with self.assertRaises(BAGError):
            raise self.err
        try:
            raise self.err
        except BAGError as e:
            self.assertIn("test", str(e))

    def test_has_message(self):
        assert hasattr(self.err, 'message')


class TestBagHelper(unittest.TestCase):

    def test_bag_samples_folder(self):
        assert os.path.exists(Helper.samples_folder())


def suite():
    s = unittest.TestSuite()
    s.addTests(unittest.TestLoader().loadTestsFromTestCase(TestBAGError))
    s.addTests(unittest.TestLoader().loadTestsFromTestCase(TestBagHelper))
    return s



import os
import logging
import argparse

from hyo2.abc2.lib.logging import set_logging
from hyo2.bag import __version__
from hyo2.bag.bag import BAGFile, is_bag

logger = logging.getLogger(__name__)


def main():
    app_name = "bag_metadata"
    app_info = "Extraction of XML metadata from an OpenNS BAG file, using hyo2.bag r%s" % __version__

    parser = argparse.ArgumentParser(prog=app_name, description=app_info)
    parser.add_argument("bag_file", type=str, help="a valid BAG file from which to extract metadata")
    parser.add_argument("-x", "--xml_file", help="the output XML metadata file", type=str)
    parser.add_argument("-v", "--verbose", help="increase output verbosity", action="store_true")
    args = parser.parse_args()

    if args.verbose:
        set_logging(ns_list=['hyo2.bag'])
        logger.debug("> verbosity: ON")

        logger.debug("> input: %s" % args.bag_file)

        if args.xml_file:
            args.xml_file = os.path.abspath(args.xml_file)
            logger.debug("> output: %s" % args.xml_file)
        else:
            args.xml_file = os.path.abspath(BAGFile.default_metadata_file)
            logger.debug("> output: %s [default]" % args.xml_file)

    if not os.path.exists(args.bag_file):
        parser.exit(1, "ERROR: the input valid does not exist: %s" % args.bag_file)

    if not is_bag(args.bag_file):
        parser.exit(1, "ERROR: the input valid does not seem a BAG file: %s" % args.bag_file)

    bf = BAGFile(args.bag_file, mode='r')
    try:
        bf.extract_metadata(args.xml_file)
    except Exception as e:
        parser.exit(1, "ERROR: %s" % e)

    if args.verbose:
        logger.debug("> DONE")


if __name__ == "__main__":
    main()



import os
import logging
from matplotlib import pyplot as plt

from hyo2.abc2.lib.logging import set_logging
from hyo2.bag.bag import BAGFile
from hyo2.bag.bag import BAGError
from hyo2.bag.helper import Helper
from hyo2.bag.uncertainty import Uncertainty2Gdal

logger = logging.getLogger(__name__)
set_logging(ns_list=['hyo2.bag'])

bag_file = os.path.join(Helper.samples_folder(), "bdb_01.bag")
if os.path.exists(bag_file):
    logger.debug("- file_bag_0: %s" % bag_file)

bag = BAGFile(bag_file, mode='r')

bag_meta = bag.populate_metadata()
logger.debug(bag_meta)

bag_uncertainty = bag.uncertainty(mask_nan=True)
logger.debug("%s: %s: %s" % (type(bag.elevation(mask_nan=True)), bag.elevation(mask_nan=True).shape,
                             bag.elevation(mask_nan=True).dtype))

Uncertainty2Gdal(bag_uncertainty=bag_uncertainty, bag_meta=bag_meta, fmt="ascii", epsg=4326)
Uncertainty2Gdal(bag_uncertainty=bag_uncertainty, bag_meta=bag_meta, fmt="geotiff", epsg=4326)
Uncertainty2Gdal(bag_uncertainty=bag_uncertainty, bag_meta=bag_meta, fmt="xyz", epsg=4326)



import os
import logging

from hyo2.abc2.lib.logging import set_logging
from hyo2.bag.bag import BAGFile
from hyo2.bag.bag import BAGError
from hyo2.bag.helper import Helper
from hyo2.bag.elevation import Elevation2Gdal

logger = logging.getLogger(__name__)
set_logging(ns_list=['hyo2.bag'])

bag_file = os.path.join(Helper.samples_folder(), "bdb_01.bag")
if os.path.exists(bag_file):
    logger.debug("- file_bag_0: %s" % bag_file)

bag = BAGFile(bag_file, mode='r')

bag_meta = bag.populate_metadata()
logger.debug(bag_meta)

bag_elevation = bag.elevation(mask_nan=False)
logger.debug(type(bag.elevation()))
logger.debug(bag.elevation().shape)
logger.debug(bag.elevation().dtype)

Elevation2Gdal(bag_elevation=bag_elevation, bag_meta=bag_meta, fmt="ascii", epsg=4326)
Elevation2Gdal(bag_elevation=bag_elevation, bag_meta=bag_meta, fmt="geotiff", epsg=4326)
Elevation2Gdal(bag_elevation=bag_elevation, bag_meta=bag_meta, fmt="xyz", epsg=4326)



import os
import logging

from hyo2.abc2.lib.logging import set_logging
from hyo2.abc2.lib.testing import Testing
from hyo2.abc2.lib.gdal_aux import GdalAux
from hyo2.bag.bag import BAGFile
from hyo2.bag.meta import Meta
from hyo2.qc4.lib.common.writers.s57_writer import S57Writer
from hyo2.qc4.lib.common.writers.kml_writer import KmlWriter
from hyo2.qc4.lib.common.writers.shp_writer import ShpWriter

set_logging(ns_list=['hyo2.bag'])
logger = logging.getLogger(__name__)

# file_bag_0 = os.path.join(Helper.samples_folder(), "bdb_01.bag")
file_bag_0 = r"D:\google_drive\_ccom\QC Tools\data\survey\BAG Checks\H13275_MB_VR_MLLW.bag"

root_folder = os.path.abspath(os.path.join(os.path.dirname(__file__), os.pardir))
testing = Testing(root_folder=root_folder)

GdalAux.push_gdal_error_handler()
GdalAux.check_gdal_data(verbose=True)

if os.path.exists(file_bag_0):
    logger.debug("- file_bag_0: %s" % file_bag_0)

bag_0 = BAGFile(file_bag_0, mode="r")
logger.debug(bag_0)

meta = Meta(bag_0.metadata())

ret = bag_0.vr_uncertainty_has_depth()
logger.debug(ret)

if not ret:
    logger.info("No flags")
    exit()

s57_bn_path = os.path.join(testing.output_data_folder(), "%s.blue_notes.000" % os.path.basename(file_bag_0))
s57_ss_path = os.path.join(testing.output_data_folder(), "%s.soundings.000" % os.path.basename(file_bag_0))

flags_for_blue_notes = list()
for entry in ret:
    flags_for_blue_notes.append([entry[0], entry[1], "%.2f" % entry[2]])

S57Writer.write_bluenotes(feature_list=flags_for_blue_notes, path=s57_bn_path, list_of_list=False)
S57Writer.write_soundings(feature_list=ret, path=s57_ss_path, list_of_list=False)
KmlWriter().write_bluenotes(feature_list=ret, path=s57_ss_path[:-4], list_of_list=False)
ShpWriter().write_bluenotes(feature_list=ret, path=s57_ss_path[:-4], list_of_list=False)



import os
import logging

from hyo2.abc2.lib.logging import set_logging
from hyo2.bag.bag import BAGFile
from hyo2.bag.helper import Helper
from hyo2.bag.meta import Meta

set_logging(ns_list=['hyo2.bag'])
logger = logging.getLogger(__name__)

# file_bag_0 = os.path.join(Helper.samples_folder(), "bdb_01.bag")
file_bag_0 = r"C:\Users\gmasetti\Desktop\H13405S_MB_8m_MLLW_1of2_edited.bag"
if os.path.exists(file_bag_0):
    logger.debug("- file_bag_0: %s" % file_bag_0)

bag_0 = BAGFile(file_bag_0, mode="r")
logger.debug(bag_0)

meta = Meta(bag_0.metadata())

output_xml = "original_metadata.xml"
bag_0.extract_metadata(output_xml)
# os.remove(output_xml)



import os
import logging
from shutil import copyfile

from hyo2.abc2.lib.logging import set_logging
from hyo2.bag.bag import BAGFile
from hyo2.bag.helper import Helper
from hyo2.bag.meta import Meta

set_logging(ns_list=['hyo2.bag'])
logger = logging.getLogger(__name__)

file_bag_0 = os.path.join(Helper.samples_folder(), "bdb_01.bag")
if os.path.exists(file_bag_0):
    logger.debug("- file_bag_0: %s" % file_bag_0)

file_bag_copy = os.path.join(os.path.dirname(__file__), "tmp_copy.bag")
bag_copy = copyfile(file_bag_0, file_bag_copy)

bag_0 = BAGFile(file_bag_copy, mode="r+")
logger.debug(bag_0)

output_xml = "fixed_metadata.xml"
if not os.path.exists(output_xml):
    raise RuntimeError("unable to find metadata file: %s" % output_xml)
bag_0.substitute_metadata(output_xml)
os.remove(output_xml)

bag_0.close()
os.remove(file_bag_copy)



import os
import logging
from shutil import copyfile
from PySide2 import QtWidgets  # required by matplotib
from matplotlib import pyplot as plt

from hyo2.abc2.lib.logging import set_logging
from hyo2.bag.bag import BAGFile
from hyo2.bag.helper import Helper

set_logging(ns_list=['hyo2.bag'])
logger = logging.getLogger(__name__)

file_bag_0 = os.path.join(Helper.samples_folder(), "bdb_01.bag")
if not os.path.exists(file_bag_0):
    raise RuntimeError("the file does not exist: %s" % file_bag_0)
logger.debug("- file_bag_0: %s" % file_bag_0)

file_bag_copy = os.path.join(os.path.dirname(__file__), "tmp_copy.bag")
bag_copy = copyfile(file_bag_0, file_bag_copy)

bag_0 = BAGFile(file_bag_copy, mode='r+')
logger.debug(bag_0)

logger.debug(type(bag_0.elevation(mask_nan=True)))
logger.debug(bag_0.elevation(mask_nan=True).shape)
logger.debug(bag_0.elevation(mask_nan=True).dtype)

ax = plt.contourf(bag_0.elevation(mask_nan=True))
plt.colorbar(ax)
plt.show()

wkt_prj_hor = """
    PROJCS["UTM Zone 19, Northern Hemisphere",
      GEOGCS["WGS 84",
        DATUM["WGS_1984",
          SPHEROID["WGS 84",6378137,298.257223563,
            AUTHORITY["EPSG","7030"]],
            TOWGS84[0,0,0,0,0,0,0],
          AUTHORITY["EPSG","6326"]],
          PRIMEM["Greenwich",0,
            AUTHORITY["EPSG","8901"]],
          UNIT["degree",0.0174532925199433,
            AUTHORITY["EPSG","9108"]],
          AUTHORITY["EPSG","4326"]],
          PROJECTION["Transverse_Mercator"],
          PARAMETER["latitude_of_origin",0],
          PARAMETER["central_meridian",-69],
          PARAMETER["scale_factor",0.9996],
          PARAMETER["false_easting",500000],
          PARAMETER["false_northing",0],
          UNIT["Meter",1]]
"""

wkt_prj_ver = """
    VERT_CS["Mean lower low water",
      VERT_DATUM["Mean lower low water",2000]]
"""

bag_0.modify_wkt_prj(wkt_prj_hor, wkt_prj_ver)
bag_0.modify_bbox(west=-70.68079657129087, east=-70.65526106943501, south=41.506684, north=41.52777)

bag_meta = bag_0.populate_metadata()
logger.debug(bag_meta)

bag_0.close()
os.remove(file_bag_copy)



import os
import logging

from osgeo import osr

from hyo2.abc2.lib.logging import set_logging
from hyo2.abc2.lib.gdal_aux import GdalAux
from hyo2.bag.bag import BAGFile
from hyo2.bag.helper import Helper
from hyo2.bag.meta import Meta

set_logging(ns_list=['hyo2.bag'])
logger = logging.getLogger(__name__)

file_bag_0 = os.path.join(Helper.samples_folder(), "bdb_01.bag")
# file_bag_0 = r""
if os.path.exists(file_bag_0):
    logger.debug("- file_bag_0: %s" % file_bag_0)

GdalAux.push_gdal_error_handler()
GdalAux.check_gdal_data()

bag_0 = BAGFile(file_bag_0, mode='r')
logger.debug(bag_0)

meta = Meta(bag_0.metadata())

logger.debug("crs: %s (%s)" % (meta.wkt_srs, type(meta.wkt_srs)))

# Try to create a transformation from BAG CRS to WGS84

osr_bag = osr.SpatialReference()
osr_bag.ImportFromWkt(meta.wkt_srs)
osr_bag.SetAxisMappingStrategy(osr.OAMS_TRADITIONAL_GIS_ORDER)
if osr_bag.IsCompound():
    osr_bag.StripVertical()
osr_geo = osr.SpatialReference()
osr_geo.ImportFromEPSG(4326)  # geographic WGS84
osr_geo.SetAxisMappingStrategy(osr.OAMS_TRADITIONAL_GIS_ORDER)
loc2geo = osr.CoordinateTransformation(osr_bag, osr_geo)

logger.debug(osr_bag.ExportToWkt())



import os
import logging
from matplotlib import pyplot as plt

from hyo2.abc2.lib.logging import set_logging
from hyo2.bag.bag import BAGFile
from hyo2.bag.bag import BAGError
from hyo2.bag.helper import Helper
from hyo2.bag.uncertainty import Uncertainty2Gdal

logger = logging.getLogger(__name__)
set_logging(ns_list=['hyo2.bag'])

bag_file = os.path.join(Helper.samples_folder(), "bdb_01.bag")
if os.path.exists(bag_file):
    logger.debug("- file_bag_0: %s" % bag_file)

bag = BAGFile(bag_file, mode='r')

bag_meta = bag.populate_metadata()
logger.debug(bag_meta)

logger.debug("has uncertainty? %s" % bag.has_uncertainty())
logger.debug("has product uncertainty? %s" % bag.has_product_uncertainty())

bag_uncertainty = bag.uncertainty(mask_nan=True)
logger.debug(type(bag.elevation(mask_nan=True)))
logger.debug(bag.elevation(mask_nan=True).shape)
logger.debug(bag.elevation(mask_nan=True).dtype)

Uncertainty2Gdal(bag_uncertainty, bag_meta, fmt="ascii")
Uncertainty2Gdal(bag_uncertainty, bag_meta, fmt="geotiff")
Uncertainty2Gdal(bag_uncertainty, bag_meta, fmt="xyz")



import os
import logging
from PySide2 import QtWidgets  # required by matplotlib
from matplotlib import pyplot as plt

from hyo2.abc2.lib.logging import set_logging
from hyo2.bag.bag import BAGFile
from hyo2.bag.bag import BAGError
from hyo2.bag.helper import Helper

logger = logging.getLogger(__name__)
set_logging(ns_list=['hyo2.bag'])

# Example that use bag.BAGFile to:
# - open a BAG file
# - read the whole elevation and uncertainty layers
# - read a selected range of rows for the elevation and uncertainty layers

file_bag_0 = os.path.join(Helper.samples_folder(), "bdb_02.bag")
if os.path.exists(file_bag_0):
    logger.debug("- file_bag_0: %s" % file_bag_0)

# - open a file
bag_0 = BAGFile(file_bag_0, mode='r')
logger.debug("\n%s\n" % bag_0)

# - get elevation shape
logger.info("elevation shape: %s" % (bag_0.elevation_shape(), ))
# - read the full elevation
full_elevation = bag_0.elevation(mask_nan=True)
logger.info("elevation array:\n  type: %s\n  shape: %s\n  dtype: %s"
            % (type(full_elevation), full_elevation.shape, full_elevation.dtype))
ax = plt.contourf(full_elevation)
plt.colorbar(ax)
plt.show()
# - read the first 10 rows of the elevation layers
selection_slice = slice(0, 10)
sliced_elevation = bag_0.elevation(mask_nan=True, row_range=selection_slice)
logger.info("sliced elevation array:\n  type: %s\n  shape: %s\n  dtype: %s"
            % (type(sliced_elevation), sliced_elevation.shape, sliced_elevation.dtype))
ax = plt.contourf(sliced_elevation)
plt.colorbar(ax)
plt.show()

# - get uncertainty shape
logger.info("uncertainty shape: %s" % (bag_0.uncertainty_shape(), ))
# - read the full uncertainty
full_uncertainty = bag_0.uncertainty(mask_nan=True)
logger.info("uncertainty array:\n  type: %s\n  shape: %s\n  dtype: %s"
            % (type(full_uncertainty), full_uncertainty.shape, full_uncertainty.dtype))
ax = plt.contourf(full_uncertainty)
plt.colorbar(ax)
plt.show()
# - read the first 10 rows of the uncertainty layers
selection_slice = slice(0, 10)
sliced_uncertainty = bag_0.uncertainty(mask_nan=True, row_range=selection_slice)
logger.info("sliced uncertainty array:\n  type: %s\n  shape: %s\n  dtype: %s"
            % (type(sliced_uncertainty), sliced_uncertainty.shape, sliced_uncertainty.dtype))
ax = plt.contourf(sliced_uncertainty)
plt.colorbar(ax)
plt.show()

# - tracking list
logger.debug("\ntracking list:\n  type: %s\n  shape: %s\n  dtype: %s"
             % (type(bag_0.tracking_list()), bag_0.tracking_list().shape, bag_0.tracking_list().dtype))

# - metadata
logger.debug("\nmetadata: %s %s\n" % (type(bag_0.metadata()), len(bag_0.metadata())))
file_bag_0_xml = os.path.join("bdb_00.bag.xml")
bag_0.extract_metadata(name=file_bag_0_xml)

bag_0.populate_metadata()
logger.debug("rows, cols: %d, %d" % (bag_0.meta.rows, bag_0.meta.cols))
logger.debug("res x, y: %f, %f" % (bag_0.meta.res_x, bag_0.meta.res_y))
logger.debug("corner SW, NE: %s, %s" % (bag_0.meta.sw, bag_0.meta.ne))
logger.debug("coord sys: %s" % bag_0.meta.wkt_srs)

logger.debug(bag_0)

file_bag_1 = os.path.join(Helper.samples_folder(), "bdb_01.bag")
if os.path.exists(file_bag_1):
    logger.debug("file_bag_1: %s" % file_bag_1)

file_bag_2 = os.path.abspath(os.path.join("test_00.bag"))
logger.debug("file_bag_2: %s" % file_bag_2)

bag_2 = BAGFile.create_template(file_bag_2)
bag_2.close()

if os.path.exists(file_bag_0_xml):
    os.remove(file_bag_0_xml)
if os.path.exists(file_bag_2):
    os.remove(file_bag_2)



import os
import logging
import numpy as np
from hyo2.bag import __version__


logger = logging.getLogger(__name__)


class TrackList2Csv(object):

    default_csv_name = "BAG.tracklist.csv"

    def __init__(self, track_list: np.ndarray, csv_file=None, header=None, comment=None):
        logger.debug("track list shape: %s" % track_list.shape)
        logger.debug("track list size: %s" % track_list.size)

        self.track_list = track_list
        if self.track_list.size == 0:
            logger.warning("nothing to export since the tracking list is empty")
            return

        self.csv_file = csv_file
        if csv_file is None:
            self.csv_file = self.default_csv_name
        self.csv_file = os.path.abspath(self.csv_file)
        logger.debug("output: %s" % self.csv_file)

        self.header = header
        if self.header is None:
            self.header = str()
        if isinstance(self.header, tuple):
            self.header = ",".join(fld for fld in self.header)
            self.header += "\n"
        logger.debug("header: %s" % self.header)

        self.comment = comment
        if self.comment is None:
            self.comment = "# Exported using BAG tools r%s\n" % __version__
        logger.debug("comment: %s" % self.comment)

        with open(self.csv_file, 'w') as f:
            f.write(self.comment)
            f.write(self.header)
            for row in track_list:
                f.write("%s, %s, %s, %s, %s, %s\n" % (row[0], row[1], row[2], row[3], row[4], row[5]))



import os
import logging

from hyo2.abc2.lib.logging import set_logging
from hyo2.bag.bag import BAGFile
from hyo2.bag.helper import Helper
from hyo2.bag.meta import Meta

set_logging(ns_list=['hyo2.bag'])
logger = logging.getLogger(__name__)

file_bag_0 = os.path.join(Helper.samples_folder(), "bdb_01.bag")
if os.path.exists(file_bag_0):
    logger.debug("- file_bag_0: %s" % file_bag_0)

bag_0 = BAGFile(file_bag_0, mode='r')
ret = bag_0.validate_metadata()
logger.debug("valid: %s" % ret)

if not ret:
    for bag_error in bag_0.meta_errors:
        logger.debug(" - %s" % bag_error)


import os
import logging
import argparse

from hyo2.abc2.lib.logging import set_logging
from hyo2.bag import __version__
from hyo2.bag.bag import BAGFile, is_bag

logger = logging.getLogger(__name__)


def main():
    app_name = "bag_tracklist"
    app_info = "Extraction the tracklist from an OpenNS BAG file, using hyo2.bag r%s" % __version__

    formats = ['csv']

    parser = argparse.ArgumentParser(prog=app_name, description=app_info)
    parser.add_argument("bag_file", type=str, help="a valid BAG file from which to extract metadata")
    parser.add_argument("-f", "--format", help="one of the available file format: " + ", ".join(formats),
                        choices=formats, default="csv", metavar='')
    parser.add_argument("-o", "--output", help="the output file", type=str)
    parser.add_argument("-hd", "--header", help="add an header", action="store_true")
    parser.add_argument("-v", "--verbose", help="increase output verbosity", action="store_true")
    args = parser.parse_args()

    if args.verbose:
        set_logging(ns_list=['hyo2.bag'])
        logger.debug("> verbosity: ON")

        logger.debug("> input: %s" % args.bag_file)

        if args.output:
            args.output = os.path.abspath(args.output)
            logger.debug("> output: %s" % args.output)
        else:
            args.output = None
            logger.debug("> output: [default]")

        logger.debug("> format: %s" % args.format)
        logger.debug("> header: %s" % args.header)

    if not os.path.exists(args.bag_file):
        parser.exit(1, "ERROR: the input valid does not exist: %s" % args.bag_file)

    if not is_bag(args.bag_file):
        parser.exit(1, "ERROR: the input valid does not seem a BAG file: %s" % args.bag_file)

    bf = BAGFile(args.bag_file, mode='r')
    tl = None
    try:
        tl = bf.tracking_list()
    except Exception as e:
        parser.exit(1, "ERROR: issue in tracking-list recovery: %s" % e)

    tlf = ""
    if args.header:
        try:
            tlf = bf.tracking_list_fields()
        except Exception as e:
            parser.exit(1, "ERROR: issue in tracking-list fields recovery: %s" % e)

    try:
        from hyo2.bag.tracklist import TrackList2Csv
        TrackList2Csv(track_list=tl, csv_file=args.output, header=tlf)
    except Exception as e:
        parser.exit(1, "ERROR: issue in output creation: %s" % e)

    if args.verbose:
        logger.debug("> DONE")


if __name__ == "__main__":
    main()



import codecs
import os
import re

from setuptools import setup, find_packages


# ------------------------------------------------------------------
#                         HELPER FUNCTIONS

here = os.path.abspath(os.path.dirname(__file__))


def read(*parts):
    # intentionally *not* adding an encoding option to open, See:
    #   https://github.com/pypa/virtualenv/issues/201#issuecomment-3145690
    with codecs.open(os.path.join(here, *parts), 'r') as fp:
        return fp.read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", version_file, re.M, )
    if version_match:
        return version_match.group(1)

    raise RuntimeError("Unable to find version string.")


# ------------------------------------------------------------------
#                          POPULATE SETUP

setup(
    name="hyo2.bag",
    version=find_version("hyo2", "bag", "__init__.py"),
    license="LGPLv3 license",

    namespace_packages=[
      "hyo2"
    ],
    packages=find_packages(exclude=[
        "*.tests", "*.tests.*", "tests.*", "tests", "*.test*",
    ]),
    package_data={
        "": ["media/*.png", "media/*.ico", "media/*.icns", "media/*.txt",],
        'hyo2.bag': [
            'iso19139/bag/*',
            'iso19139/gco/*',
            'iso19139/gmd/*',
            'iso19139/gmi/*',
            'iso19139/gml/*.xsd',
            'iso19139/gml/*.txt',
            'iso19139/gml/3.1.1/smil/*',
            'iso19139/gsr/*',
            'iso19139/gss/*',
            'iso19139/gts/*',
            'iso19139/xlink/*',
            'iso19757-3/*',
            'samples/*',
        ],
    },
    zip_safe=False,
    setup_requires=[
        "setuptools",
        "wheel",
    ],
    install_requires=[
        "hyo2.abc2>=2.3.7",
        "python-dateutil",
        "lxml",
        "numpy",
        "gdal",
        "h5py"
    ],
    python_requires='>=3.8',
    entry_points={
        "gui_scripts": [
        ],
        "console_scripts": [
            'bag_bbox = hyo2.bag.tools.bag_bbox:main',
            'bag_elevation = hyo2.bag.tools.bag_elevation:main',
            'bag_metadata = hyo2.bag.tools.bag_metadata:main',
            'bag_tracklist = hyo2.bag.tools.bag_tracklist:main',
            'bag_uncertainty = hyo2.bag.tools.bag_uncertainty:main',
            'bag_validate = hyo2.bag.tools.bag_validate:main'
        ],
    },
    test_suite="tests",

    description="A package to manage Bathymetric Attributed Grid (BAG) data files.",
    long_description=(read('README.rst') + '\n\n\"\"\"\"\"\"\"\n\n' +
                      read('HISTORY.rst') + '\n\n\"\"\"\"\"\"\"\n\n' +
                      read('AUTHORS.rst') + '\n\n\"\"\"\"\"\"\"\n\n' +
                      read(os.path.join('docs', 'how_to_contribute.rst')))
,
    url="https://www.hydroffice.org/bag",
    classifiers=[  # https://pypi.python.org/pypi?%3Aaction=list_classifiers
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Science/Research',
        'Natural Language :: English',
        'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Topic :: Scientific/Engineering :: GIS',
        'Topic :: Office/Business :: Office Suites',
    ],
    keywords="hydrography ocean mapping survey bag openns",
    author="Giuseppe Masetti (CCOM,UNH)",
    author_email="gmasetti@ccom.unh.edu",
)



import logging
import dateutil.parser
from lxml import etree
from osgeo import osr
from hyo2.abc2.lib.gdal_aux import GdalAux
from hyo2.bag.helper import Helper

logger = logging.getLogger(__name__)


class Meta:
    """ Helper class to manage BAG xml metadata. """

    ns = {
        'bag': 'http://www.opennavsurf.org/schema/bag',
        'gco': 'http://www.isotc211.org/2005/gco',
        'gmd': 'http://www.isotc211.org/2005/gmd',
        'gmi': 'http://www.isotc211.org/2005/gmi',
        'gml': 'http://www.opengis.net/gml/3.2',
        'xsi': 'http://www.w3.org/2001/XMLSchema-instance',
    }

    ns2 = {
        'gml': 'http://www.opengis.net/gml',
        'xsi': 'http://www.w3.org/2001/XMLSchema-instance',
        'smXML': 'http://metadata.dgiwg.org/smXML',
    }

    def __init__(self, meta_xml: bytes | str) -> None:
        GdalAux.push_gdal_error_handler()

        self.xml_tree = etree.fromstring(meta_xml)

        # rows and cols
        self.rows: int | None = None
        self.cols: int | None = None
        self._read_rows_and_cols()

        # resolution along x and y axes
        self.res_x: float | None = None
        self.res_y: float | None = None
        self._read_res_x_and_y()

        # corner SW and NE
        self.sw: list[float] | None = None
        self.ne: list[float] | None = None
        self._read_corners_sw_and_ne()

        # wkt projection
        self.wkt_srs: str | None = None
        self.xml_srs: str | None = None
        self.wkt_srs_epsg_code: str | None = None
        self._read_wkt_prj()

        # wkt vertical datum
        self.wkt_vertical_datum: str | None = None
        self.xml_vertical_datum: str | None = None
        self.wkt_vertical_datum_epsg_code: str | None = None
        self._read_wkt_vertical_datum()

        # bbox
        self.lon_min: float | None = None
        self.lon_max: float | None = None
        self.lat_min: float | None = None
        self.lat_max: float | None = None
        self._read_bbox()

        # abstract
        self.abstract = None
        self._read_abstract()

        # date
        self.date = None
        self._read_date()

        # survey dates
        self.survey_start_date = None
        self._read_survey_start_date()
        self.survey_end_date = None
        self._read_survey_end_date()

        # uncertainty type
        self.unc_type = None
        self._read_uncertainty_type()

        # uncertainty type
        self.sec_constr = None
        self._read_security_constraints()

    def __str__(self) -> str:
        output = "<metadata>"

        if (self.rows is not None) and (self.cols is not None):
            output += "\n    <shape rows=%d, cols=%d>" % (self.rows, self.cols)

        if (self.res_x is not None) and (self.res_y is not None):
            output += "\n    <resolution x=%f, y=%f>" % (self.res_x, self.res_y)

        if (self.sw is not None) and (self.ne is not None):
            output += "\n    <corners SW=%s, NE=%s>" % (self.sw, self.ne)

        if self.wkt_srs is not None:
            output += "\n    <projection=%s>" % Helper.elide(self.wkt_srs, max_len=60)

        if self.date is not None:
            output += "\n    <date=%s>" % self.date

        if self.abstract is not None:
            output += "\n    <abstract=%s>" % self.abstract

        output += "\n    <bbox>"
        if (self.lon_min is not None) and (self.lon_max is not None):
            output += "\n        <x min=%s, max=%s>" % (self.lon_min, self.lon_max)
        if (self.lat_min is not None) and (self.lat_max is not None):
            output += "\n        <y min=%s, max=%s>" % (self.lat_min, self.lat_max)

        if self.unc_type is not None:
            output += "\n    <uncertainty type=%s>" % self.unc_type

        if self.sec_constr is not None:
            output += "\n    <security constraints=%s>" % self.sec_constr

        return output

    def valid_bbox(self) -> bool:
        return (self.lon_min is not None) and (self.lon_max is not None) and \
               (self.lat_min is not None) and (self.lat_max is not None)

    def geo_extent(self) -> tuple[float, float, float, float]:
        """ Return the geographic extent as a tuple: (x_min, x_max, y_min, y_max) """
        return self.lon_min, self.lon_max, self.lat_min, self.lat_max

    def wkt_bbox(self) -> str:
        return "LINESTRING Z(%.6f %.6f 0, %.6f %.6f 0, %.6f %.6f 0, %.6f %.6f 0, %.6f %.6f 0)" \
               % (self.lon_min, self.lat_min, self.lon_min, self.lat_max, self.lon_max, self.lat_max, self.lon_max,
                  self.lat_min, self.lon_min, self.lat_min)

    def _read_rows_and_cols(self) -> None:
        """ attempts to read rows and cols info """

        try:
            ret = self.xml_tree.xpath('//*/gmd:spatialRepresentationInfo/gmd:MD_Georectified/'
                                      'gmd:axisDimensionProperties/gmd:MD_Dimension/gmd:dimensionSize/gco:Integer',
                                      namespaces=self.ns)
        except etree.Error as e:
            logger.warning("unable to read rows and cols: %s" % e)
            return

        if len(ret) == 0:

            try:
                ret = self.xml_tree.xpath('//*/spatialRepresentationInfo/smXML:MD_Georectified/'
                                          'axisDimensionProperties/smXML:MD_Dimension/dimensionSize',
                                          namespaces=self.ns2)
            except etree.Error as e:
                logger.warning("unable to read rows and cols: %s" % e)
                return

        try:
            self.rows = int(ret[0].text)
            self.cols = int(ret[1].text)

        except (ValueError, IndexError) as e:
            logger.warning("unable to read rows and cols: %s" % e)
            return

    def _read_res_x_and_y(self) -> None:
        """ attempts to read resolution along x- and y- axes """

        try:
            ret = self.xml_tree.xpath('//*/gmd:spatialRepresentationInfo/gmd:MD_Georectified/'
                                      'gmd:axisDimensionProperties/gmd:MD_Dimension/gmd:resolution/gco:Measure',
                                      namespaces=self.ns)
        except etree.Error as e:
            logger.warning("unable to read res x and y: %s" % e)
            return

        if len(ret) == 0:

            try:
                ret = self.xml_tree.xpath('//*/spatialRepresentationInfo/smXML:MD_Georectified/'
                                          'axisDimensionProperties/smXML:MD_Dimension/resolution/'
                                          'smXML:Measure/smXML:value',
                                          namespaces=self.ns2)
            except etree.Error as e:
                logger.warning("unable to read res x and y: %s" % e)
                return

        try:
            self.res_x = float(ret[0].text)
            self.res_y = float(ret[1].text)

        except (ValueError, IndexError) as e:
            logger.warning("unable to read res x and y: %s" % e)
            return

    def _read_corners_sw_and_ne(self) -> None:
        """ attempts to read corners SW and NE """

        try:
            ret = self.xml_tree.xpath('//*/gmd:spatialRepresentationInfo/gmd:MD_Georectified/'
                                      'gmd:cornerPoints/gml:Point/gml:coordinates',
                                      namespaces=self.ns)[0].text.split()
        except (etree.Error, IndexError):
            try:
                ret = self.xml_tree.xpath('//*/spatialRepresentationInfo/smXML:MD_Georectified/'
                                          'cornerPoints/gml:Point/gml:coordinates',
                                          namespaces=self.ns2)[0].text.split()
            except (etree.Error, IndexError) as e:
                logger.warning("unable to read corners SW and NE: %s" % e)
                return

        try:
            self.sw = [float(c) for c in ret[0].split(',')]
            self.ne = [float(c) for c in ret[1].split(',')]

        except (ValueError, IndexError) as e:
            logger.warning("unable to read corners SW and NE: %s" % e)
            return

    def _read_wkt_prj(self) -> None:
        """ attempts to read the WKT projection string """

        try:
            ret = self.xml_tree.xpath('//*/gmd:referenceSystemInfo/gmd:MD_ReferenceSystem/'
                                      'gmd:referenceSystemIdentifier/gmd:RS_Identifier/gmd:code/gco:CharacterString',
                                      namespaces=self.ns)
        except etree.Error as e:
            logger.warning("unable to read the WKT projection string: %s" % e)
            return

        if len(ret) == 0:
            try:
                ret = self.xml_tree.xpath('//*/referenceSystemInfo/smXML:MD_CRS',
                                          namespaces=self.ns2)
            except etree.Error as e:
                logger.warning("unable to read the WKT projection string: %s" % e)
                return

            if len(ret) != 0:
                logger.warning("unsupported method to describe CRS")
                self.xml_srs = etree.tostring(ret[0], pretty_print=True)
                # print(self.xml_srs)
                return

        try:
            space = self.xml_tree.xpath(
                '//*/gmd:referenceSystemInfo/gmd:MD_ReferenceSystem/'
                'gmd:referenceSystemIdentifier/gmd:RS_Identifier/gmd:codeSpace/gco:CharacterString',
                namespaces=self.ns)
            # logger.info("codeSpace: %s" % space[0].text)

            if space[0].text == "EPSG":
                self.wkt_srs_epsg_code = int(ret[0].text)
                sr = osr.SpatialReference()
                sr.ImportFromEPSG(self.wkt_srs_epsg_code)
                self.wkt_srs = sr.ExportToWkt()
            else:
                self.wkt_srs_epsg_code = None
                self.wkt_srs = ret[0].text

        except (ValueError, IndexError) as e:
            logger.warning("unable to read the WKT projection string: %s" % e)
            return

    def _read_wkt_vertical_datum(self) -> None:
        """ attempts to read the WKT vertical datum string """

        try:
            ret = self.xml_tree.xpath('//*/gmd:referenceSystemInfo/gmd:MD_ReferenceSystem/'
                                      'gmd:referenceSystemIdentifier/gmd:RS_Identifier/gmd:code/gco:CharacterString',
                                      namespaces=self.ns)
        except etree.Error as e:
            logger.warning("unable to read the WKT vertical datum string: %s" % e)
            return

        if len(ret) == 0:
            try:
                ret = self.xml_tree.xpath('//*/referenceSystemInfo/smXML:MD_CRS',
                                          namespaces=self.ns2)
            except etree.Error as e:
                logger.warning("unable to read the WKT vertical datum string: %s" % e)
                return

            if len(ret) != 0:
                logger.warning("unsupported method to describe Vertical Datum")
                self.xml_vertical_datum = etree.tostring(ret[0], pretty_print=True)
                # print(self.xml_vertical_datum)
                return

        try:
            space = self.xml_tree.xpath(
                '//*/gmd:referenceSystemInfo/gmd:MD_ReferenceSystem/'
                'gmd:referenceSystemIdentifier/gmd:RS_Identifier/gmd:codeSpace/gco:CharacterString',
                namespaces=self.ns)
            # logger.info("codeSpace: %s" % space[0].text)

            if space[1].text == "EPSG":
                self.wkt_vertical_datum_epsg_code = int(ret[1].text)
                sr = osr.SpatialReference()
                sr.ImportFromEPSG(self.wkt_vertical_datum_epsg_code)
                self.wkt_vertical_datum = sr.ExportToWkt()
            else:
                self.wkt_vertical_datum_epsg_code = None
                self.wkt_vertical_datum = ret[1].text

        except (ValueError, IndexError) as e:
            logger.warning("unable to read the WKT vertical datum string: %s" % e)
            return

    def _read_bbox(self) -> None:
        """ attempts to read the bounding box values """

        try:
            ret_x_min = self.xml_tree.xpath('//*/gmd:EX_GeographicBoundingBox/gmd:westBoundLongitude/gco:Decimal',
                                            namespaces=self.ns)
            ret_x_max = self.xml_tree.xpath('//*/gmd:EX_GeographicBoundingBox/gmd:eastBoundLongitude/gco:Decimal',
                                            namespaces=self.ns)
        except etree.Error as e:
            logger.warning("unable to read the bbox's longitude values: %s" % e)
            return

        if len(ret_x_min) == 0:
            try:
                ret_x_min = self.xml_tree.xpath('//*/smXML:EX_GeographicBoundingBox/westBoundLongitude',
                                                namespaces=self.ns2)
                ret_x_max = self.xml_tree.xpath('//*/smXML:EX_GeographicBoundingBox/eastBoundLongitude',
                                                namespaces=self.ns2)
            except etree.Error as e:
                logger.warning("unable to read the bbox's longitude values: %s" % e)
                return

        try:
            self.lon_min = float(ret_x_min[0].text)
            self.lon_max = float(ret_x_max[0].text)
        except (ValueError, IndexError) as e:
            logger.warning("unable to read the bbox's longitude values: %s" % e)
            return

        try:
            ret_y_min = self.xml_tree.xpath('//*/gmd:EX_GeographicBoundingBox/gmd:southBoundLatitude/gco:Decimal',
                                            namespaces=self.ns)
            ret_y_max = self.xml_tree.xpath('//*/gmd:EX_GeographicBoundingBox/gmd:northBoundLatitude/gco:Decimal',
                                            namespaces=self.ns)
        except etree.Error as e:
            logger.warning("unable to read the bbox's latitude values: %s" % e)
            return

        if len(ret_y_min) == 0:

            try:
                ret_y_min = self.xml_tree.xpath('//*/smXML:EX_GeographicBoundingBox/southBoundLatitude',
                                                namespaces=self.ns2)
                ret_y_max = self.xml_tree.xpath('//*/smXML:EX_GeographicBoundingBox/northBoundLatitude',
                                                namespaces=self.ns2)
            except etree.Error as e:
                logger.warning("unable to read the bbox's latitude values: %s" % e)
                return

        try:
            self.lat_min = float(ret_y_min[0].text)
            self.lat_max = float(ret_y_max[0].text)
        except (ValueError, IndexError) as e:
            logger.warning("unable to read the bbox's latitude values: %s" % e)
            return

    def _read_abstract(self) -> None:
        """ attempts to read the abstract string """

        try:
            ret = self.xml_tree.xpath('//*/gmd:abstract/gco:CharacterString',
                                      namespaces=self.ns)
        except etree.Error as e:
            logger.warning("unable to read the abstract string: %s" % e)
            return

        if len(ret) == 0:
            try:
                ret = self.xml_tree.xpath('//*/abstract',
                                          namespaces=self.ns2)
            except etree.Error as e:
                logger.warning("unable to read the abstract string: %s" % e)
                return

        try:
            self.abstract = ret[0].text
        except (ValueError, IndexError) as e:
            logger.warning("unable to read the abstract string: %s" % e)
            return

    def _read_date(self) -> None:
        """ attempts to read the date string """

        ret = self.xml_tree.xpath('//*/gmd:CI_Date/gmd:date/gco:Date',
                                  namespaces=self.ns)

        if len(ret) == 0:

            ret = self.xml_tree.xpath('//*/smXML:CI_Date/date',
                                      namespaces=self.ns2)

        if len(ret) == 0:

            ret = self.xml_tree.xpath('//*/gmd:dateStamp/gco:Date',
                                      namespaces=self.ns)

        if len(ret) == 0:
            logger.warning("unable to read the date string")

        try:
            text_date = ret[0].text
        except (ValueError, IndexError) as e:
            logger.warning("unable to read the date string: %s" % e)
            return

        tm_date = None
        # noinspection PyBroadException
        try:
            parsed_date = dateutil.parser.parse(text_date)
            tm_date = parsed_date.strftime('%Y-%m-%dT%H:%M:%SZ')
        except Exception as e:
            logger.warning("unable to handle the date string: %s (%s)" % (text_date, e), exc_info=True)

        if tm_date is None:
            self.date = text_date
        else:
            self.date = tm_date

    def _read_survey_start_date(self) -> None:
        """ attempts to read the survey date strings """

        try:
            ret_begin = self.xml_tree.xpath(
                '//*/gmd:identificationInfo/bag:BAG_DataIdentification/gmd:extent/gmd:EX_Extent/'
                'gmd:temporalElement/gmd:EX_TemporalExtent/gmd:extent/gml:TimePeriod/gml:beginPosition',
                namespaces=self.ns)

            if len(ret_begin) == 0:
                ret_begin = self.xml_tree.xpath(
                    '//*/identificationInfo/smXML:BAG_DataIdentification/extent/smXML:EX_Extent/'
                    'temporalElement/smXML:EX_TemporalExtent/extent/TimePeriod/beginPosition',
                    namespaces=self.ns2)

        except Exception as e:
            logger.warning(e, exc_info=True)
            return

        if len(ret_begin) == 0:
            logger.warning("unable to read the survey begin date string")
            return

        try:
            text_begin_date = ret_begin[0].text
        except (ValueError, IndexError) as e:
            logger.warning("unable to read the survey begin date string: %s" % e)
            return

        tm_begin_date = None
        # noinspection PyBroadException
        try:
            parsed_date = dateutil.parser.parse(text_begin_date)
            tm_begin_date = parsed_date.strftime('%Y-%m-%dT%H:%M:%SZ')
        except Exception as e:
            logger.warning("unable to handle the survey begin date string: %s (%s)" % (text_begin_date, e),
                           exc_info=True)

        if tm_begin_date is None:
            self.survey_start_date = text_begin_date
        else:
            self.survey_start_date = tm_begin_date

        # logger.debug('start: %s' % self.survey_start_date)

    def _read_survey_end_date(self) -> None:
        """ attempts to read the survey date strings """

        try:
            ret_end = self.xml_tree.xpath(
                '//*/gmd:identificationInfo/bag:BAG_DataIdentification/gmd:extent/gmd:EX_Extent/'
                'gmd:temporalElement/gmd:EX_TemporalExtent/gmd:extent/gml:TimePeriod/gml:endPosition',
                namespaces=self.ns)

            if len(ret_end) == 0:
                ret_end = self.xml_tree.xpath(
                    '//*/identificationInfo/smXML:BAG_DataIdentification/extent/smXML:EX_Extent/'
                    'temporalElement/smXML:EX_TemporalExtent/extent/TimePeriod/endPosition',
                    namespaces=self.ns2)
        except Exception as e:
            logger.warning(e, exc_info=True)
            return

        if len(ret_end) == 0:
            logger.warning("unable to read the survey end date string")
            return

        try:
            text_end_date = ret_end[0].text
        except (ValueError, IndexError) as e:
            logger.warning("unable to read the survey end date string: %s" % e)
            return

        tm_end_date = None
        # noinspection PyBroadException
        try:
            parsed_date = dateutil.parser.parse(text_end_date)
            tm_end_date = parsed_date.strftime('%Y-%m-%dT%H:%M:%SZ')
        except Exception as e:
            logger.warning("unable to handle the survey end date string: %s (%s)" % (text_end_date, e), exc_info=True)

        if tm_end_date is None:
            self.survey_end_date = text_end_date
        else:
            self.survey_end_date = tm_end_date

        # logger.debug('end: %s' % self.survey_end_date)

    def _read_uncertainty_type(self) -> None:
        """ attempts to read the uncertainty type """
        old_format = False

        try:
            ret = self.xml_tree.xpath('//*/bag:verticalUncertaintyType/bag:BAG_VertUncertCode/@codeListValue',
                                      namespaces=self.ns)
        except etree.Error as e:
            logger.warning("unable to read the uncertainty type string: %s" % e)
            return

        if len(ret) == 0:

            try:
                ret = self.xml_tree.xpath('//*/verticalUncertaintyType',
                                          namespaces=self.ns2)
                old_format = True
            except etree.Error as e:
                logger.warning("unable to read the uncertainty type string: %s" % e)
                return

        try:
            if old_format:
                self.unc_type = ret[0].text
            else:
                self.unc_type = ret[0]
        except (ValueError, IndexError) as e:
            logger.warning("unable to read the uncertainty type attribute: %s" % e)
            return

    def _read_security_constraints(self) -> None:
        """ attempts to read the uncertainty type """
        old_format = False

        try:
            ret = self.xml_tree.xpath(
                '//*/gmd:MD_SecurityConstraints/gmd:classification/gmd:MD_ClassificationCode/@codeListValue',
                namespaces=self.ns)
        except etree.Error as e:
            logger.warning("unable to read the uncertainty type string: %s" % e)
            return

        if len(ret) == 0:

            try:
                ret = self.xml_tree.xpath('//*/smXML:MD_SecurityConstraints/classification',
                                          namespaces=self.ns2)
                old_format = True
            except etree.Error as e:
                logger.warning("unable to read the uncertainty type string: %s" % e)
                return

        try:
            if old_format:
                self.sec_constr = ret[0].text
            else:
                self.sec_constr = ret[0]
        except (ValueError, IndexError) as e:
            logger.warning("unable to read the uncertainty type attribute: %s" % e)
            return



"""
Hydro-Package
BAG
"""

name = 'BAG'
__version__ = '1.2.12'
__author__ = 'gmasetti@ccom.unh.edu'
__license__ = 'LGPLv3 license'
__copyright__ = 'Copyright (c) 2025, University of New Hampshire, Center for Coastal and Ocean Mapping'



import os
import logging

from osgeo import ogr, osr
from hyo2.bag.meta import Meta
from hyo2.bag.helper import BAGError, Helper
from hyo2.bag import __version__


logger = logging.getLogger(__name__)
ogr.UseExceptions()


class Bbox2Gdal(object):

    formats = {
        'gjs': ["GeoJSON", "bag.geojson"],
        'gml': ["GML", "bag.gml"],
        'kml': ["KML", "bag.kml"],
        'shp': ["ESRI Shapefile", "bag.shp"],
    }

    def __init__(self, bag_meta: Meta, fmt="kml", title=None, out_file=None):
        self.bag_meta = bag_meta
        if not self.bag_meta.valid_bbox():
            raise BAGError("invalid bbox read in BAG metadata")

        self.title = title
        if self.title is None:
            self.title = "Metadata"
        logger.debug("title: %s" % self.title)

        # get the ogr driver
        self.drv = ogr.GetDriverByName(self.formats[fmt][0])
        if self.drv is None:
            raise BAGError("%s driver not available.\n" % self.formats[fmt][0])

        # set the output file
        self.out_file = out_file
        if self.out_file is None:
            self.out_file = os.path.abspath(self.formats[fmt][1])
            logger.debug("output: %s" % self.out_file)

        if os.path.exists(self.out_file):
            os.remove(self.out_file)

        # create the data source
        ds = self.drv.CreateDataSource(self.out_file)

        # create the spatial reference (WGS84)
        self.srs = osr.SpatialReference()
        self.srs.ImportFromEPSG(4326)

        # create the layer
        self.lyr = ds.CreateLayer("BAG", self.srs, ogr.wkbLineString25D)
        self._define_layer_fields()

        # add feature
        self._add_feature()

        ds.Destroy()

    def _define_layer_fields(self):

        # Add the fields we're interested in
        self.lyr.CreateField(ogr.FieldDefn("Name", ogr.OFTString))
        if self.bag_meta.rows is not None:
            self.lyr.CreateField(ogr.FieldDefn("Rows", ogr.OFTInteger))
        if self.bag_meta.cols is not None:
            self.lyr.CreateField(ogr.FieldDefn("Cols", ogr.OFTInteger))
        if self.bag_meta.ne is not None:
            self.lyr.CreateField(ogr.FieldDefn("NE", ogr.OFTString))
        if self.bag_meta.sw is not None:
            self.lyr.CreateField(ogr.FieldDefn("SW", ogr.OFTString))
        if self.bag_meta.res_x is not None:
            self.lyr.CreateField(ogr.FieldDefn("ResX", ogr.OFTReal))
        if self.bag_meta.res_y is not None:
            self.lyr.CreateField(ogr.FieldDefn("ResY", ogr.OFTReal))
        if self.bag_meta.abstract is not None:
            self.lyr.CreateField(ogr.FieldDefn("Abstract", ogr.OFTString))
        if self.bag_meta.date is not None:
            self.lyr.CreateField(ogr.FieldDefn("Date", ogr.OFTString))
        if self.bag_meta.wkt_srs is not None:
            self.lyr.CreateField(ogr.FieldDefn("SRS", ogr.OFTString))
        self.lyr.CreateField(ogr.FieldDefn("Tools", ogr.OFTString))

    def _add_feature(self):
        # create the WKT for the feature using Python string formatting
        feature = ogr.Feature(self.lyr.GetLayerDefn())
        feature.SetField("Name", self.title)
        if self.bag_meta.rows is not None:
            feature.SetField("Rows", self.bag_meta.rows)
        if self.bag_meta.cols is not None:
            feature.SetField("Cols", self.bag_meta.cols)
        if self.bag_meta.ne is not None:
            feature.SetField("NE", ("%s" % self.bag_meta.ne))
        if self.bag_meta.sw is not None:
            feature.SetField("SW", ("%s" % self.bag_meta.sw))
        if self.bag_meta.res_x is not None:
            feature.SetField("ResX", ("%s" % self.bag_meta.res_x))
        if self.bag_meta.res_y is not None:
            feature.SetField("ResY", ("%s" % self.bag_meta.res_y))
        if self.bag_meta.abstract is not None:
            feature.SetField("Abstract", self.bag_meta.abstract)
        if self.bag_meta.date is not None:
            feature.SetField("Date", self.bag_meta.date)
        if self.bag_meta.wkt_srs is not None:
            feature.SetField("SRS", Helper.elide(self.bag_meta.wkt_srs, max_len=60))
        feature.SetField("Tools", ("r%s" % __version__))
        wkt = self.bag_meta.wkt_bbox()
        # log.debug("bbox: %s" % wkt)
        point = ogr.CreateGeometryFromWkt(wkt)
        feature.SetGeometry(point)
        self.lyr.CreateFeature(feature)
        feature.Destroy()
